"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.pkgutil as pkgutil
import salt.utils.pkg
from salt.exceptions import CommandExecutionError, MinionError
from tests.support.mock import MagicMock, Mock, patch


@pytest.fixture
def configure_loader_modules():
    return {pkgutil: {}}


# 'refresh_db' function tests: 1


def test_refresh_db():
    """
    Test if it updates the pkgutil repo database (pkgutil -U).
    """
    mock = MagicMock(return_value=0)
    with patch.dict(pkgutil.__salt__, {"cmd.retcode": mock}):
        with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
            assert pkgutil.refresh_db()


# 'upgrade_available' function tests: 1


def test_upgrade_available():
    """
    Test if there is an upgrade available for a certain package.
    """
    mock = MagicMock(return_value="A\n B\n SAME")
    with patch.dict(pkgutil.__salt__, {"cmd.run_stdout": mock}):
        assert pkgutil.upgrade_available("CSWpython") == ""

    mock = MagicMock(side_effect=["A\n B\n SALT", None])
    with patch.dict(pkgutil.__salt__, {"cmd.run_stdout": mock}):
        assert pkgutil.upgrade_available("CSWpython") == "SALT"

        assert pkgutil.upgrade_available("CSWpython") == ""


# 'list_upgrades' function tests: 1


def test_list_upgrades():
    """
    Test if it list all available package upgrades on this system.
    """
    mock_run = MagicMock(return_value="A\t B\t SAME")
    mock_ret = MagicMock(return_value=0)
    with patch.dict(
        pkgutil.__salt__, {"cmd.run_stdout": mock_run, "cmd.retcode": mock_ret}
    ):
        with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
            assert pkgutil.list_upgrades() == {"A": " B"}


# 'upgrade' function tests: 1


def test_upgrade():
    """
    Test if it upgrade all of the packages to the latest available version.
    """
    mock_run = MagicMock(return_value="A\t B\t SAME")
    mock_ret = MagicMock(return_value=0)
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        pkgutil.__salt__,
        {
            "cmd.run_stdout": mock_run,
            "cmd.retcode": mock_ret,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run_all": mock_ret,
            "cmd.run": mock_run,
        },
    ):
        with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": mock_ret}):
            with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
                assert pkgutil.upgrade() == {}


# 'list_pkgs' function tests: 1


def test_list_pkgs():
    """
    Test if it list the packages currently installed as a dict.
    """
    mock_run = MagicMock(return_value="A\t B\t SAME")
    mock_ret = MagicMock(return_value=True)
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        pkgutil.__salt__,
        {
            "cmd.run_stdout": mock_run,
            "cmd.retcode": mock_ret,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run": mock_run,
        },
    ):
        with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": mock_ret}):
            assert pkgutil.list_pkgs(versions_as_list=True, removed=True) == {}

        assert pkgutil.list_pkgs() == {}

    with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": True}):
        assert pkgutil.list_pkgs(versions_as_list=True)

        mock_pkg = MagicMock(return_value=True)
        with patch.dict(pkgutil.__salt__, {"pkg_resource.stringify": mock_pkg}):
            assert pkgutil.list_pkgs()


def test_list_pkgs_no_context():
    """
    Test if it list the packages currently installed as a dict.
    """
    mock_run = MagicMock(return_value="A\t B\t SAME")
    mock_ret = MagicMock(return_value=True)
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        pkgutil.__salt__,
        {
            "cmd.run_stdout": mock_run,
            "cmd.retcode": mock_ret,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run": mock_run,
        },
    ), patch.object(pkgutil, "_list_pkgs_from_context") as list_pkgs_context_mock:
        pkgs = pkgutil.list_pkgs(versions_as_list=True, removed=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()

        pkgs = pkgutil.list_pkgs(versions_as_list=True, removed=True, use_context=False)
        list_pkgs_context_mock.assert_not_called()
        list_pkgs_context_mock.reset_mock()


# 'version' function tests: 1


def test_version():
    """
    Test if it returns a version if the package is installed.
    """
    mock_ret = MagicMock(return_value=True)
    with patch.dict(pkgutil.__salt__, {"pkg_resource.version": mock_ret}):
        assert pkgutil.version("CSWpython")


# 'latest_version' function tests: 1


def test_latest_version():
    """
    Test if it return the latest version of the named package
    available for upgrade or installation.
    """
    assert pkgutil.latest_version() == ""

    mock_run_all = MagicMock(return_value="A\t B\t SAME")
    mock_run = MagicMock(return_value={"stdout": ""})
    mock_ret = MagicMock(return_value=True)
    mock_pkg = MagicMock(return_value="")
    with patch.dict(
        pkgutil.__salt__,
        {
            "cmd.retcode": mock_ret,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run_all": mock_run,
            "cmd.run": mock_run_all,
        },
    ):
        with patch.object(salt.utils.pkg, "clear_rtag", Mock()):
            assert pkgutil.latest_version("CSWpython") == ""

            assert pkgutil.latest_version("CSWpython", "Python") == {
                "Python": "",
                "CSWpython": "",
            }


# 'install' function tests: 1


def test_install():
    """
    Test if it install packages using the pkgutil tool.
    """
    mock_pkg = MagicMock(side_effect=MinionError)
    with patch.dict(pkgutil.__salt__, {"pkg_resource.parse_targets": mock_pkg}):
        pytest.raises(CommandExecutionError, pkgutil.install)

    mock_ret = MagicMock(return_value=True)
    mock_pkg = MagicMock(return_value=[""])
    with patch.dict(pkgutil.__salt__, {"pkg_resource.parse_targets": mock_pkg}):
        with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": mock_ret}):
            assert pkgutil.install() == {}

    mock_run = MagicMock(return_value="A\t B\t SAME")
    mock_run_all = MagicMock(return_value={"stdout": ""})
    mock_pkg = MagicMock(return_value=[{"bar": "1.2.3"}])
    with patch.dict(
        pkgutil.__salt__,
        {
            "pkg_resource.parse_targets": mock_pkg,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run_all": mock_run_all,
            "cmd.run": mock_run,
        },
    ):
        with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": mock_ret}):
            assert pkgutil.install(pkgs='["foo", {"bar": "1.2.3"}]') == {}


# 'remove' function tests: 1


def test_remove():
    """
    Test if it remove a package and all its dependencies
    which are not in use by other packages.
    """
    mock_pkg = MagicMock(side_effect=MinionError)
    with patch.dict(pkgutil.__salt__, {"pkg_resource.parse_targets": mock_pkg}):
        pytest.raises(CommandExecutionError, pkgutil.remove)

    mock_ret = MagicMock(return_value=True)
    mock_run = MagicMock(return_value="A\t B\t SAME")
    mock_run_all = MagicMock(return_value={"stdout": ""})
    mock_pkg = MagicMock(return_value=[""])
    with patch.dict(
        pkgutil.__salt__,
        {
            "pkg_resource.parse_targets": mock_pkg,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run_all": mock_run_all,
            "cmd.run": mock_run,
        },
    ):
        with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": mock_ret}):
            assert pkgutil.remove() == {}

    mock_pkg = MagicMock(return_value=[{"bar": "1.2.3"}])
    with patch.dict(
        pkgutil.__salt__,
        {
            "pkg_resource.parse_targets": mock_pkg,
            "pkg_resource.stringify": mock_pkg,
            "pkg_resource.sort_pkglist": mock_pkg,
            "cmd.run_all": mock_run_all,
            "cmd.run": mock_run,
        },
    ):
        with patch.dict(pkgutil.__context__, {"pkg.list_pkgs": mock_ret}):
            with patch.object(pkgutil, "list_pkgs", return_value={"bar": "1.2.3"}):
                assert pkgutil.remove(pkgs='["foo", "bar"]') == {}


# 'purge' function tests: 1


def test_purge():
    """
    Test if it package purges are not supported,
    this function is identical to ``remove()``.
    """
    mock_pkg = MagicMock(side_effect=MinionError)
    with patch.dict(pkgutil.__salt__, {"pkg_resource.parse_targets": mock_pkg}):
        pytest.raises(CommandExecutionError, pkgutil.purge)



import urllib.parse

import pytest

import salt.utils.json
from salt.netapi.rest_tornado import saltnado
from tests.support.mock import MagicMock, patch


@pytest.fixture
def app_urls():
    return [
        (r"/hook(/.*)?", saltnado.WebhookSaltAPIHandler),
    ]


async def test_hook_can_handle_get_parameters(http_client, app, content_type_map):
    with patch("salt.utils.event.get_event") as get_event:
        with patch.dict(app.mod_opts, {"webhook_disable_auth": True}):
            event = MagicMock()
            event.fire_event.return_value = True
            get_event.return_value = event
            response = await http_client.fetch(
                "/hook/my_service/?param=1&param=2",
                body=salt.utils.json.dumps({}),
                method="POST",
                headers={"Content-Type": content_type_map["json"]},
            )
            assert response.code == 200
            host = urllib.parse.urlparse(response.effective_url).netloc
            event.fire_event.assert_called_once_with(
                {
                    "headers": {
                        "Content-Length": "2",
                        "Connection": "close",
                        "Content-Type": "application/json",
                        "Host": host,
                        "Accept-Encoding": "gzip",
                    },
                    "post": {},
                    "get": {"param": ["1", "2"]},
                },
                "salt/netapi/hook/my_service/",
            )



"""
NAPALM YANG
===========

NAPALM YANG basic operations.

.. versionadded:: 2017.7.0
"""

import logging

# import NAPALM utils
import salt.utils.napalm
from salt.utils.napalm import proxy_napalm_wrap

try:
    import napalm_yang

    HAS_NAPALM_YANG = True
except ImportError:
    HAS_NAPALM_YANG = False


# -----------------------------------------------------------------------------
# module properties
# -----------------------------------------------------------------------------

__virtualname__ = "napalm_yang"
__proxyenabled__ = ["*"]
# uses NAPALM-based proxy to interact with network devices

log = logging.getLogger(__file__)

# -----------------------------------------------------------------------------
# property functions
# -----------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    This module in particular requires also napalm-yang.
    """
    if not HAS_NAPALM_YANG:
        return (
            False,
            "Unable to load napalm_yang execution module: please install napalm-yang!",
        )
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# -----------------------------------------------------------------------------
# helper functions -- will not be exported
# -----------------------------------------------------------------------------


def _get_root_object(models):
    """
    Read list of models and returns a Root object with the proper models added.
    """
    root = napalm_yang.base.Root()
    for model in models:
        current = napalm_yang
        for part in model.split("."):
            current = getattr(current, part)
        root.add_model(current)
    return root


# -----------------------------------------------------------------------------
# callable functions
# -----------------------------------------------------------------------------


def diff(candidate, running, *models):
    """
    Returns the difference between two configuration entities structured
    according to the YANG model.

    .. note::
        This function is recommended to be used mostly as a state helper.

    candidate
        First model to compare.

    running
        Second model to compare.

    models
        A list of models to be used when comparing.

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_yang.diff {} {} models.openconfig_interfaces

    Output Example:

    .. code-block:: python

        {
            "interfaces": {
                "interface": {
                    "both": {
                        "Port-Channel1": {
                            "config": {
                                "mtu": {
                                    "first": "0",
                                    "second": "9000"
                                }
                            }
                        }
                    },
                    "first_only": [
                        "Loopback0"
                    ],
                    "second_only": [
                        "Loopback1"
                    ]
                }
            }
        }
    """
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]

    first = _get_root_object(models)
    first.load_dict(candidate)
    second = _get_root_object(models)
    second.load_dict(running)
    return napalm_yang.utils.diff(first, second)


@proxy_napalm_wrap
def parse(*models, **kwargs):
    """
    Parse configuration from the device.

    models
        A list of models to be used when parsing.

    config: ``False``
        Parse config.

    state: ``False``
        Parse state.

    profiles: ``None``
        Use certain profiles to parse. If not specified, will use the device
        default profile(s).

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_yang.parse models.openconfig_interfaces

    Output Example:

    .. code-block:: python

        {
            "interfaces": {
                "interface": {
                    ".local.": {
                        "name": ".local.",
                        "state": {
                            "admin-status": "UP",
                            "counters": {
                                "in-discards": 0,
                                "in-errors": 0,
                                "out-errors": 0
                            },
                            "enabled": True,
                            "ifindex": 0,
                            "last-change": 0,
                            "oper-status": "UP",
                            "type": "softwareLoopback"
                        },
                        "subinterfaces": {
                            "subinterface": {
                                ".local..0": {
                                    "index": ".local..0",
                                    "state": {
                                        "ifindex": 0,
                                        "name": ".local..0"
                                    }
                                }
                            }
                        }
                    },
                    "ae0": {
                        "name": "ae0",
                        "state": {
                            "admin-status": "UP",
                            "counters": {
                                "in-discards": 0,
                                "in-errors": 0,
                                "out-errors": 0
                            },
                            "enabled": True,
                            "ifindex": 531,
                            "last-change": 255203,
                            "mtu": 1518,
                            "oper-status": "DOWN"
                        },
                        "subinterfaces": {
                            "subinterface": {
                                "ae0.0": {
                                    "index": "ae0.0",
                                    "state": {
                                        "description": "ASDASDASD",
                                        "ifindex": 532,
                                        "name": "ae0.0"
                                    }
                                }
                                "ae0.32767": {
                                    "index": "ae0.32767",
                                    "state": {
                                        "ifindex": 535,
                                        "name": "ae0.32767"
                                    }
                                }
                            }
                        }
                    },
                    "dsc": {
                        "name": "dsc",
                        "state": {
                            "admin-status": "UP",
                            "counters": {
                                "in-discards": 0,
                                "in-errors": 0,
                                "out-errors": 0
                            },
                            "enabled": True,
                            "ifindex": 5,
                            "last-change": 0,
                            "oper-status": "UP"
                        }
                    },
                    "ge-0/0/0": {
                        "name": "ge-0/0/0",
                        "state": {
                            "admin-status": "UP",
                            "counters": {
                                "in-broadcast-pkts": 0,
                                "in-discards": 0,
                                "in-errors": 0,
                                "in-multicast-pkts": 0,
                                "in-unicast-pkts": 16877,
                                "out-broadcast-pkts": 0,
                                "out-errors": 0,
                                "out-multicast-pkts": 0,
                                "out-unicast-pkts": 15742
                            },
                            "description": "management interface",
                            "enabled": True,
                            "ifindex": 507,
                            "last-change": 258467,
                            "mtu": 1400,
                            "oper-status": "UP"
                        },
                        "subinterfaces": {
                            "subinterface": {
                                "ge-0/0/0.0": {
                                    "index": "ge-0/0/0.0",
                                    "state": {
                                        "description": "ge-0/0/0.0",
                                        "ifindex": 521,
                                        "name": "ge-0/0/0.0"
                                    }
                                }
                            }
                        }
                    }
                    "irb": {
                        "name": "irb",
                        "state": {
                            "admin-status": "UP",
                            "counters": {
                                "in-discards": 0,
                                "in-errors": 0,
                                "out-errors": 0
                            },
                            "enabled": True,
                            "ifindex": 502,
                            "last-change": 0,
                            "mtu": 1514,
                            "oper-status": "UP",
                            "type": "ethernetCsmacd"
                        }
                    },
                    "lo0": {
                        "name": "lo0",
                        "state": {
                            "admin-status": "UP",
                            "counters": {
                                "in-discards": 0,
                                "in-errors": 0,
                                "out-errors": 0
                            },
                            "description": "lo0",
                            "enabled": True,
                            "ifindex": 6,
                            "last-change": 0,
                            "oper-status": "UP",
                            "type": "softwareLoopback"
                        },
                        "subinterfaces": {
                            "subinterface": {
                                "lo0.0": {
                                    "index": "lo0.0",
                                    "state": {
                                        "description": "lo0.0",
                                        "ifindex": 16,
                                        "name": "lo0.0"
                                    }
                                },
                                "lo0.16384": {
                                    "index": "lo0.16384",
                                    "state": {
                                        "ifindex": 21,
                                        "name": "lo0.16384"
                                    }
                                },
                                "lo0.16385": {
                                    "index": "lo0.16385",
                                    "state": {
                                        "ifindex": 22,
                                        "name": "lo0.16385"
                                    }
                                },
                                "lo0.32768": {
                                    "index": "lo0.32768",
                                    "state": {
                                        "ifindex": 248,
                                        "name": "lo0.32768"
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    """
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]
    config = kwargs.pop("config", False)
    state = kwargs.pop("state", False)
    profiles = kwargs.pop("profiles", [])
    # pylint: disable=undefined-variable
    if not profiles and hasattr(napalm_device, "profile"):
        profiles = napalm_device.profile  # pylint: disable=undefined-variable
    # pylint: enable=undefined-variable
    if not profiles:
        profiles = [__grains__.get("os")]
    root = _get_root_object(models)
    parser_kwargs = {
        "device": napalm_device.get("DRIVER"),  # pylint: disable=undefined-variable
        "profile": profiles,
    }
    if config:
        root.parse_config(**parser_kwargs)
    if state:
        root.parse_state(**parser_kwargs)
    return root.to_dict(filter=True)


@proxy_napalm_wrap
def get_config(data, *models, **kwargs):
    """
    Return the native config.

    data
        Dictionary structured with respect to the models referenced.

    models
        A list of models to be used when generating the config.

    profiles: ``None``
        Use certain profiles to generate the config.
        If not specified, will use the platform default profile(s).

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_yang.get_config {} models.openconfig_interfaces

    Output Example:

    .. code-block:: text

        interface et1
            ip address 192.168.1.1/24
            description Uplink1
            mtu 9000
        interface et2
            ip address 192.168.2.1/24
            description Uplink2
            mtu 9000
    """
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]
    profiles = kwargs.pop("profiles", [])
    # pylint: disable=undefined-variable
    if not profiles and hasattr(napalm_device, "profile"):
        # pylint: enable=undefined-variable
        profiles = napalm_device.profile  # pylint: disable=undefined-variable
    if not profiles:
        profiles = [__grains__.get("os")]
    parser_kwargs = {"profile": profiles}
    root = _get_root_object(models)
    root.load_dict(data)
    native_config = root.translate_config(**parser_kwargs)
    log.debug("Generated config")
    log.debug(native_config)
    return native_config


@proxy_napalm_wrap
def load_config(data, *models, **kwargs):
    """
    Generate and load the config on the device using the OpenConfig or IETF
    models and device profiles.

    data
        Dictionary structured with respect to the models referenced.

    models
        A list of models to be used when generating the config.

    profiles: ``None``
        Use certain profiles to generate the config.
        If not specified, will use the platform default profile(s).

    test: ``False``
        Dry run? If set as ``True``, will apply the config, discard
        and return the changes. Default: ``False`` and will commit
        the changes on the device.

    commit: ``True``
        Commit? Default: ``True``.

    debug: ``False``
        Debug mode. Will insert a new key under the output dictionary,
        as ``loaded_config`` containing the raw configuration loaded on the device.

    replace: ``False``
        Should replace the config with the new generate one?

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_yang.load_config {} models.openconfig_interfaces test=True debug=True

    Output Example:

    .. code-block:: jinja

        device1:
            ----------
            already_configured:
                False
            comment:
            diff:
                [edit interfaces ge-0/0/0]
                -   mtu 1400;
                [edit interfaces ge-0/0/0 unit 0 family inet]
                -       dhcp;
                [edit interfaces lo0]
                -    unit 0 {
                -        description lo0.0;
                -    }
                +    unit 1 {
                +        description "new loopback";
                +    }
            loaded_config:
                <configuration>
                  <interfaces replace="replace">
                    <interface>
                      <name>ge-0/0/0</name>
                      <unit>
                        <name>0</name>
                        <family>
                          <inet/>
                        </family>
                        <description>ge-0/0/0.0</description>
                      </unit>
                      <description>management interface</description>
                    </interface>
                    <interface>
                      <name>ge-0/0/1</name>
                      <disable/>
                      <description>ge-0/0/1</description>
                    </interface>
                    <interface>
                      <name>ae0</name>
                      <unit>
                        <name>0</name>
                        <vlan-id>100</vlan-id>
                        <family>
                          <inet>
                            <address>
                              <name>192.168.100.1/24</name>
                            </address>
                            <address>
                              <name>172.20.100.1/24</name>
                            </address>
                          </inet>
                        </family>
                        <description>a description</description>
                      </unit>
                      <vlan-tagging/>
                      <unit>
                        <name>1</name>
                        <vlan-id>1</vlan-id>
                        <family>
                          <inet>
                            <address>
                              <name>192.168.101.1/24</name>
                            </address>
                          </inet>
                        </family>
                        <disable/>
                        <description>ae0.1</description>
                      </unit>
                      <vlan-tagging/>
                      <unit>
                        <name>2</name>
                        <vlan-id>2</vlan-id>
                        <family>
                          <inet>
                            <address>
                              <name>192.168.102.1/24</name>
                            </address>
                          </inet>
                        </family>
                        <description>ae0.2</description>
                      </unit>
                      <vlan-tagging/>
                    </interface>
                    <interface>
                      <name>lo0</name>
                      <unit>
                        <name>1</name>
                        <description>new loopback</description>
                      </unit>
                      <description>lo0</description>
                    </interface>
                  </interfaces>
                </configuration>
            result:
                True
    """
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]
    config = get_config(data, *models, **kwargs)
    test = kwargs.pop("test", False)
    debug = kwargs.pop("debug", False)
    commit = kwargs.pop("commit", True)
    replace = kwargs.pop("replace", False)
    # pylint: disable=undefined-variable
    return __salt__["net.load_config"](
        text=config,
        test=test,
        debug=debug,
        commit=commit,
        replace=replace,
        inherit_napalm_device=napalm_device,
    )
    # pylint: enable=undefined-variable


@proxy_napalm_wrap
def compliance_report(data, *models, **kwargs):
    """
    Return the compliance report using YANG objects.

    data
        Dictionary structured with respect to the models referenced.

    models
        A list of models to be used when generating the config.

    filepath
        The absolute path to the validation file.

    CLI Example:

    .. code-block:: bash

        salt '*' napalm_yang.compliance_report {} models.openconfig_interfaces filepath=~/validate.yml

    Output Example:

    .. code-block:: json

        {
          "skipped": [],
          "complies": true,
          "get_interfaces_ip": {
            "missing": [],
            "complies": true,
            "present": {
              "ge-0/0/0.0": {
                "complies": true,
                "nested": true
              }
            },
            "extra": []
          }
        }
    """
    if isinstance(models, tuple) and isinstance(models[0], list):
        models = models[0]
    filepath = kwargs.pop("filepath", "")
    root = _get_root_object(models)
    root.load_dict(data)
    return root.compliance_report(filepath)



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.modules.postfix as postfix
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {postfix: {}}


def test_show_master():
    """
    Test for return a dict of active config values
    """
    with patch.object(postfix, "_parse_master", return_value=({"A": "a"}, ["b"])):
        assert postfix.show_master("path") == {"A": "a"}


def test_set_master():
    """
    Test for set a single config value in the master.cf file
    """
    with patch.object(postfix, "_parse_master", return_value=({"A": "a"}, ["b"])):
        with patch.object(postfix, "_write_conf", return_value=None):
            assert postfix.set_master("a", "b")


def test_show_main():
    """
    Test for return a dict of active config values
    """
    with patch.object(postfix, "_parse_main", return_value=({"A": "a"}, ["b"])):
        assert postfix.show_main("path") == {"A": "a"}


def test_set_main():
    """
    Test for set a single config value in the master.cf file
    """
    with patch.object(postfix, "_parse_main", return_value=({"A": "a"}, ["b"])):
        with patch.object(postfix, "_write_conf", return_value=None):
            assert postfix.set_main("key", "value")


def test_show_queue():
    """
    Test for show contents of the mail queue
    """
    with patch.dict(postfix.__salt__, {"cmd.run": MagicMock(return_value="A\nB")}):
        assert postfix.show_queue() == []

    # Test if get an extra newline in the output
    with patch.dict(postfix.__salt__, {"cmd.run": MagicMock(return_value="A\nB\n")}):
        assert postfix.show_queue() == []


def test_delete():
    """
    Test for delete message(s) from the mail queue
    """
    with patch.object(postfix, "show_queue", return_value={}):
        assert postfix.delete("queue_id") == {
            "result": False,
            "message": "No message in queue with ID queue_id",
        }

    with patch.dict(
        postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert postfix.delete("ALL") == {
            "result": True,
            "message": "Successfully removed all messages",
        }


def test_hold():
    """
    Test for set held message(s) in the mail queue to unheld
    """
    with patch.object(postfix, "show_queue", return_value={}):
        assert postfix.hold("queue_id") == {
            "result": False,
            "message": "No message in queue with ID queue_id",
        }

    with patch.dict(
        postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert postfix.hold("ALL") == {
            "result": True,
            "message": "Successfully placed all messages on hold",
        }


def test_unhold():
    """
    Test for put message(s) on hold from the mail queue
    """
    with patch.object(postfix, "show_queue", return_value={}):
        assert postfix.unhold("queue_id") == {
            "result": False,
            "message": "No message in queue with ID queue_id",
        }

    with patch.dict(
        postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert postfix.unhold("ALL") == {
            "result": True,
            "message": "Successfully set all message as unheld",
        }


def test_requeue():
    """
    Test for requeue message(s) in the mail queue
    """
    with patch.object(postfix, "show_queue", return_value={}):
        assert postfix.requeue("queue_id") == {
            "result": False,
            "message": "No message in queue with ID queue_id",
        }

    with patch.dict(
        postfix.__salt__, {"cmd.run_all": MagicMock(return_value={"retcode": 0})}
    ):
        assert postfix.requeue("ALL") == {
            "result": True,
            "message": "Successfully requeued all messages",
        }



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os
import shutil
import uuid
from collections import OrderedDict

import pytest

import salt.modules.seed as seed
import salt.utils.files
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {seed: {}}


@pytest.mark.slow_test
def test_mkconfig_odict():
    with patch.dict(seed.__opts__, {"master": "foo"}):
        ddd = OrderedDict()
        ddd["b"] = "b"
        ddd["a"] = "b"
        data = seed.mkconfig(ddd, approve_key=False)
        with salt.utils.files.fopen(data["config"]) as fic:
            fdata = fic.read()
            assert fdata == "b: b\na: b\nmaster: foo\n"


def test_prep_bootstrap():
    """
    Test to update and get the random script to a random place
    """
    with patch.dict(
        seed.__salt__,
        {
            "config.gather_bootstrap_script": MagicMock(
                return_value=os.path.join("BS_PATH", "BS")
            )
        },
    ), patch.object(uuid, "uuid4", return_value="UUID"), patch.object(
        os.path, "exists", return_value=True
    ), patch.object(
        os, "chmod", return_value=None
    ), patch.object(
        shutil, "copy", return_value=None
    ):

        expect = (
            os.path.join("MPT", "tmp", "UUID", "BS"),
            os.sep + os.path.join("tmp", "UUID"),
        )
        assert seed.prep_bootstrap("MPT") == expect

        expect = (
            os.sep + os.path.join("MPT", "tmp", "UUID", "BS"),
            os.sep + os.path.join("tmp", "UUID"),
        )
        assert seed.prep_bootstrap(os.sep + "MPT") == expect


def test_apply_():
    """
    Test to seed a location (disk image, directory, or block device)
    with the minion config, approve the minion's key, and/or install
    salt-minion.
    """
    mock = MagicMock(
        side_effect=[
            False,
            {"type": "type", "target": "target"},
            {"type": "type", "target": "target"},
            {"type": "type", "target": "target"},
        ]
    )
    with patch.dict(seed.__salt__, {"file.stats": mock}):
        assert seed.apply_("path") == "path does not exist"

        with patch.object(seed, "_mount", return_value=False):
            assert seed.apply_("path") == "target could not be mounted"

        with patch.object(seed, "_mount", return_value="/mountpoint"):
            with patch.object(os.path, "join", return_value="A"):
                with patch.object(os, "makedirs", MagicMock(side_effect=OSError("f"))):
                    with patch.object(os.path, "isdir", return_value=False):
                        pytest.raises(OSError, seed.apply_, "p")

                with patch.object(os, "makedirs", MagicMock()):
                    with patch.object(seed, "mkconfig", return_value="A"):
                        with patch.object(seed, "_check_install", return_value=False):
                            with patch.object(
                                seed, "_umount", return_value=None
                            ) as umount_mock:
                                assert not seed.apply_("path", install=False)
                                umount_mock.assert_called_once_with(
                                    "/mountpoint", "target", "type"
                                )



import pytest

import salt.states.saltmod as saltmod
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(minion_opts):
    return {
        saltmod: {
            "__env__": "base",
            "__opts__": minion_opts,
        },
    }


def test_test_mode():
    name = "bah"

    expected = {
        "name": name,
        "changes": {},
        "result": None,
        "comment": f"Wheel function '{name}' would be executed.",
    }

    with patch.dict(saltmod.__opts__, {"test": True}):
        ret = saltmod.wheel(name)
        assert ret == expected


def test_wheel():
    """
    Test to execute a wheel module on the master
    """
    name = "state"

    expected = {
        "changes": {"return": True},
        "name": "state",
        "result": True,
        "comment": "Wheel function 'state' executed.",
    }
    with patch.dict(
        saltmod.__salt__, {"saltutil.wheel": MagicMock(return_value={"return": True})}
    ):
        ret = saltmod.wheel(name)
        assert ret == expected


def test_test_error_in_return():
    name = "bah"

    jid = "20170406104341210934"
    func_ret = {"Error": "This is an Error!"}
    expected = {
        "name": name,
        "changes": {"return": func_ret},
        "result": False,
        "comment": f"Wheel function '{name}' failed.",
        "__jid__": jid,
    }

    mock = MagicMock(return_value={"return": func_ret, "jid": jid})
    with patch.dict(saltmod.__salt__, {"saltutil.wheel": mock}):
        ret = saltmod.wheel(name)
        assert ret == expected
        mock.assert_called_once()



def test_pillar_using_http_query(salt_master, salt_minion, salt_cli):
    pillar_top = """
    base:
      "*":
        - http_pillar_test
    """
    my_pillar = """
    {%- set something = salt['http.query']('https://raw.githubusercontent.com/saltstack/salt/master/.pre-commit-config.yaml', raise_error=False, verify_ssl=False, status=True, timeout=15).status %}
    http_query_test: {{ something }}
    """

    with salt_master.pillar_tree.base.temp_file("top.sls", pillar_top):
        with salt_master.pillar_tree.base.temp_file("http_pillar_test.sls", my_pillar):

            # We may need this for the following pillar.item to work
            ret = salt_cli.run("saltutil.pillar_refresh", minion_tgt=salt_minion.id)
            assert ret.returncode == 0

            pillar_ret = salt_cli.run(
                "pillar.item", "http_query_test", minion_tgt=salt_minion.id
            )
            assert pillar_ret.returncode == 0
            assert '"http_query_test": 200' in pillar_ret.stdout



import os
import threading
import time

import pytest

import salt.auth
import salt.ext.tornado.escape
import salt.ext.tornado.web
import salt.utils.json
import salt.utils.stringutils
from salt.ext.tornado.testing import AsyncHTTPTestCase
from salt.netapi.rest_tornado import saltnado
from tests.support.helpers import TstSuiteLoggingHandler, patched_environ
from tests.support.mixins import AdaptedConfigurationTestCaseMixin


@pytest.fixture(scope="module", autouse=True)
def salt_api_account(salt_api_account_factory):
    with salt_api_account_factory as account:
        yield account


class SaltnadoIntegrationTestsBase(
    AsyncHTTPTestCase, AdaptedConfigurationTestCaseMixin
):

    content_type_map = {
        "json": "application/json",
        "json-utf8": "application/json; charset=utf-8",
        "yaml": "application/x-yaml",
        "text": "text/plain",
        "form": "application/x-www-form-urlencoded",
        "xml": "application/xml",
        "real-accept-header-json": "application/json, text/javascript, */*; q=0.01",
        "real-accept-header-yaml": "application/x-yaml, text/yaml, */*; q=0.01",
    }
    auth_creds = (
        ("username", "saltdev_api"),
        ("password", "saltdev"),
        ("eauth", "auto"),
    )

    @property
    def auth_creds_dict(self):
        return dict(self.auth_creds)

    @property
    def opts(self):
        return self.get_temp_config("client_config")

    @property
    def mod_opts(self):
        return self.get_temp_config("minion")

    @property
    def auth(self):
        if not hasattr(self, "__auth"):
            self.__auth = salt.auth.LoadAuth(self.opts)
        return self.__auth

    @property
    def token(self):
        """Mint and return a valid token for auth_creds"""
        return self.auth.mk_token(self.auth_creds_dict)

    def setUp(self):
        super().setUp()
        self.patched_environ = patched_environ(ASYNC_TEST_TIMEOUT="30")
        self.patched_environ.__enter__()  # pylint: disable=unnecessary-dunder-call
        self.addCleanup(self.patched_environ.__exit__)

    def tearDown(self):
        super().tearDown()
        if hasattr(self, "http_server"):
            del self.http_server
        if hasattr(self, "io_loop"):
            del self.io_loop
        if hasattr(self, "_app"):
            del self._app
        if hasattr(self, "http_client"):
            del self.http_client
        if hasattr(self, "__port"):
            del self.__port
        if hasattr(self, "_AsyncHTTPTestCase__port"):
            del self._AsyncHTTPTestCase__port
        if hasattr(self, "__auth"):
            del self.__auth
        if hasattr(self, "_SaltnadoIntegrationTestsBase__auth"):
            del self._SaltnadoIntegrationTestsBase__auth
        if hasattr(self, "_test_generator"):
            del self._test_generator
        if hasattr(self, "application"):
            del self.application
        if hasattr(self, "patched_environ"):
            del self.patched_environ

    def build_tornado_app(self, urls):
        application = salt.ext.tornado.web.Application(urls, debug=True)

        application.auth = self.auth
        application.opts = self.opts
        application.mod_opts = self.mod_opts

        return application

    def decode_body(self, response):
        if response is None:
            return response
        if response.body:
            # Decode it
            if response.headers.get("Content-Type") == "application/json":
                response._body = response.body.decode("utf-8")
            else:
                response._body = salt.ext.tornado.escape.native_str(response.body)
        return response

    def fetch(self, path, **kwargs):
        return self.decode_body(super().fetch(path, **kwargs))

    def get_app(self):
        raise NotImplementedError


@pytest.mark.usefixtures("salt_sub_minion")
class TestSaltAPIHandler(SaltnadoIntegrationTestsBase):
    def setUp(self):
        super().setUp()
        os.environ["ASYNC_TEST_TIMEOUT"] = "300"

    def get_app(self):
        urls = [("/", saltnado.SaltAPIHandler)]

        application = self.build_tornado_app(urls)

        application.event_listener = saltnado.EventListener({}, self.opts)
        self.application = application
        return application

    @pytest.mark.slow_test
    def test_regression_49572(self):
        with TstSuiteLoggingHandler() as handler:
            GATHER_JOB_TIMEOUT = 1
            self.application.opts["gather_job_timeout"] = GATHER_JOB_TIMEOUT

            low = [{"client": "local", "tgt": "*", "fun": "test.ping"}]
            fetch_kwargs = {
                "method": "POST",
                "body": salt.utils.json.dumps(low),
                "headers": {
                    "Content-Type": self.content_type_map["json"],
                    saltnado.AUTH_TOKEN_HEADER: self.token["token"],
                },
                "connect_timeout": 30,
                "request_timeout": 30,
            }

            self.fetch("/", **fetch_kwargs)
            time.sleep(GATHER_JOB_TIMEOUT + 0.1)  # ick

            #  While the traceback is in the logs after the sleep without this
            #  follow up fetch, the logging handler doesn't see it in its list
            #  of messages unless something else runs.
            self.fetch("/", **fetch_kwargs)

            for message in handler.messages:
                if "TypeError: 'NoneType' object is not iterable" in message:
                    raise AssertionError(
                        "#49572: regression: set_result on completed event"
                    )

    @pytest.mark.skip(
        reason="Undetermined race condition in test. Temporarily disabled."
    )
    def test_simple_local_post_only_dictionary_request_with_order_masters(self):
        """
        Test a basic API of /
        """
        low = {
            "client": "local",
            "tgt": "*",
            "fun": "test.ping",
        }

        self.application.opts["order_masters"] = True
        self.application.opts["syndic_wait"] = 5

        response = self.fetch(
            "/",
            method="POST",
            body=salt.utils.json.dumps(low),
            headers={
                "Content-Type": self.content_type_map["json"],
                saltnado.AUTH_TOKEN_HEADER: self.token["token"],
            },
            connect_timeout=30,
            request_timeout=30,
        )
        response_obj = salt.utils.json.loads(response.body)
        self.application.opts["order_masters"] = []
        self.application.opts["syndic_wait"] = 5
        # If --proxy is set, it will cause an extra minion_id to be in the
        # response. Since there's not a great way to know if the test runner's
        # proxy minion is running, and we're not testing proxy minions here
        # anyway, just remove it from the response.
        response_obj[0]["return"].pop("proxytest", None)
        self.assertEqual(response_obj["return"], [{"minion": True, "sub_minion": True}])


class TestWebhookSaltAPIHandler(SaltnadoIntegrationTestsBase):
    def get_app(self):

        urls = [
            (r"/hook(/.*)?", saltnado.WebhookSaltAPIHandler),
        ]

        application = self.build_tornado_app(urls)

        self.application = application

        application.event_listener = saltnado.EventListener({}, self.opts)
        return application

    @pytest.mark.skip(
        reason="Skipping until we can devote more resources to debugging this test."
    )
    def test_post(self):
        self._future_resolved = threading.Event()
        try:

            def verify_event(future):
                """
                Notify the threading event that the future is resolved
                """
                self._future_resolved.set()

            self._finished = (
                False  # TODO: remove after some cleanup of the event listener
            )

            # get an event future
            future = self.application.event_listener.get_event(
                self, tag="salt/netapi/hook", callback=verify_event
            )
            # fire the event
            response = self.fetch(
                "/hook",
                method="POST",
                body="foo=bar",
                headers={saltnado.AUTH_TOKEN_HEADER: self.token["token"]},
            )
            response_obj = salt.utils.json.loads(response.body)
            self.assertTrue(response_obj["success"])
            resolve_future_timeout = 60
            self._future_resolved.wait(resolve_future_timeout)
            try:
                event = future.result()
            except Exception as exc:  # pylint: disable=broad-except
                self.fail(
                    "Failed to resolve future under {} secs: {}".format(
                        resolve_future_timeout, exc
                    )
                )
            self.assertEqual(event["tag"], "salt/netapi/hook")
            self.assertIn("headers", event["data"])
            self.assertEqual(event["data"]["post"], {"foo": "bar"})
        finally:
            self._future_resolved.clear()
            del self._future_resolved



import errno
import subprocess

import salt.utils.systemd as _systemd
from salt.exceptions import SaltInvocationError
from tests.support.mock import Mock, patch
from tests.support.unit import TestCase


def _booted_effect(path):
    return path == "/run/systemd/system"


def _not_booted_effect(path):
    if path == "/run/systemd/system":
        raise OSError(errno.ENOENT, "No such file or directory", path)


class SystemdTestCase(TestCase):
    """
    Tests the functions in salt.utils.systemd
    """

    def test_booted(self):
        """
        Test that salt.utils.systemd.booted() returns True when minion is
        systemd-booted.
        """
        # Ensure that os.stat returns True. os.stat doesn't return a bool
        # normally, but the code is doing a simple truth check on the return
        # data, so it is sufficient enough to mock it as True for these tests.
        with patch("os.stat", side_effect=_booted_effect):
            # Test without context dict passed
            self.assertTrue(_systemd.booted())
            # Test that context key is set when context dict is passed
            context = {}
            self.assertTrue(_systemd.booted(context))
            self.assertEqual(context, {"salt.utils.systemd.booted": True})

    def test_not_booted(self):
        """
        Test that salt.utils.systemd.booted() returns False when minion is not
        systemd-booted.
        """
        # Ensure that os.stat raises an exception even if test is being run on
        # a systemd-booted host.
        with patch("os.stat", side_effect=_not_booted_effect):
            # Test without context dict passed
            self.assertFalse(_systemd.booted())
            # Test that context key is set when context dict is passed
            context = {}
            self.assertFalse(_systemd.booted(context))
            self.assertEqual(context, {"salt.utils.systemd.booted": False})

    def test_booted_return_from_context(self):
        """
        Test that the context data is returned when present. To ensure we're
        getting data from the context dict, we use a non-boolean value to
        differentiate it from the True/False return this function normally
        produces.
        """
        context = {"salt.utils.systemd.booted": "foo"}
        self.assertEqual(_systemd.booted(context), "foo")

    def test_booted_invalid_context(self):
        """
        Test with invalid context data. The context value must be a dict, so
        this should raise a SaltInvocationError.
        """
        # Test with invalid context data
        with self.assertRaises(SaltInvocationError):
            _systemd.booted(99999)

    def test_version(self):
        """
        Test that salt.utils.systemd.booted() returns True when minion is
        systemd-booted.
        """
        with patch("subprocess.Popen") as popen_mock:
            _version = 231
            output = f"systemd {_version}\n-SYSVINIT"
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: (output, None),
                pid=lambda: 12345,
                retcode=0,
            )

            # Test without context dict passed
            self.assertEqual(_systemd.version(), _version)
            # Test that context key is set when context dict is passed
            context = {}
            self.assertTrue(_systemd.version(context))
            self.assertEqual(context, {"salt.utils.systemd.version": _version})

    def test_version_generated_from_git_describe(self):
        """
        Test with version string matching versions generated by git describe
        in systemd. This feature is used in systemd>=241.
        """
        with patch("subprocess.Popen") as popen_mock:
            _version = 241
            output = "systemd {0} ({0}.0-0-dist)\n-SYSVINIT".format(_version)
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: (output, None),
                pid=lambda: 12345,
                retcode=0,
            )

            # Test without context dict passed
            self.assertEqual(_systemd.version(), _version)
            # Test that context key is set when context dict is passed
            context = {}
            self.assertTrue(_systemd.version(context))
            self.assertEqual(context, {"salt.utils.systemd.version": _version})

    def test_version_return_from_context(self):
        """
        Test that the context data is returned when present. To ensure we're
        getting data from the context dict, we use a non-integer value to
        differentiate it from the integer return this function normally
        produces.
        """
        context = {"salt.utils.systemd.version": "foo"}
        self.assertEqual(_systemd.version(context), "foo")

    def test_version_invalid_context(self):
        """
        Test with invalid context data. The context value must be a dict, so
        this should raise a SaltInvocationError.
        """
        # Test with invalid context data
        with self.assertRaises(SaltInvocationError):
            _systemd.version(99999)

    def test_version_parse_problem(self):
        """
        Test with invalid context data. The context value must be a dict, so
        this should raise a SaltInvocationError.
        """
        with patch("subprocess.Popen") as popen_mock:
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: ("invalid", None),
                pid=lambda: 12345,
                retcode=0,
            )
            # Test without context dict passed
            self.assertIsNone(_systemd.version())
            # Test that context key is set when context dict is passed. A failure
            # to parse the systemctl output should not set a context key, so it
            # should not be present in the context dict.
            context = {}
            self.assertIsNone(_systemd.version(context))
            self.assertEqual(context, {})

    def test_has_scope_systemd204(self):
        """
        Scopes are available in systemd>=205. Make sure that this function
        returns the expected boolean. We do three separate unit tests for
        versions 204 through 206 because mock doesn't like us altering the
        return_value in a loop.
        """
        with patch("subprocess.Popen") as popen_mock:
            _expected = False
            _version = 204
            _output = f"systemd {_version}\n-SYSVINIT"
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: (_output, None),
                pid=lambda: 12345,
                retcode=0,
            )
            # Ensure that os.stat returns True. os.stat doesn't return a bool
            # normally, but the code is doing a simple truth check on the
            # return data, so it is sufficient enough to mock it as True for
            # these tests.
            with patch("os.stat", side_effect=_booted_effect):
                with patch("salt.utils.systemd.status", return_value=True):
                    # Test without context dict passed
                    self.assertEqual(_systemd.has_scope(), _expected)
                    context = {"salt.utils.systemd.status": True}
                    # Test that context key is set when context dict is passed
                    self.assertEqual(_systemd.has_scope(context), _expected)
                    self.assertEqual(
                        context,
                        {
                            "salt.utils.systemd.booted": True,
                            "salt.utils.systemd.status": True,
                            "salt.utils.systemd.version": _version,
                        },
                    )

    def test_has_scope_systemd205(self):
        """
        Scopes are available in systemd>=205. Make sure that this function
        returns the expected boolean. We do three separate unit tests for
        versions 204 through 206 because mock doesn't like us altering the
        return_value in a loop.
        """
        with patch("subprocess.Popen") as popen_mock:
            _expected = True
            _version = 205
            _output = f"systemd {_version}\n-SYSVINIT"
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: (_output, None),
                pid=lambda: 12345,
                retcode=0,
            )
            # Ensure that os.stat returns True. os.stat doesn't return a bool
            # normally, but the code is doing a simple truth check on the
            # return data, so it is sufficient enough to mock it as True for
            # these tests.
            with patch("os.stat", side_effect=_booted_effect):
                with patch("salt.utils.systemd.status", return_value=True):
                    # Test without context dict passed
                    self.assertEqual(_systemd.has_scope(), _expected)
                    # Test that context key is set when context dict is passed
                    context = {"salt.utils.systemd.status": True}
                    self.assertEqual(_systemd.has_scope(context), _expected)
                    self.assertEqual(
                        context,
                        {
                            "salt.utils.systemd.booted": True,
                            "salt.utils.systemd.version": _version,
                            "salt.utils.systemd.status": True,
                        },
                    )

    def test_has_scope_systemd206(self):
        """
        Scopes are available in systemd>=205. Make sure that this function
        returns the expected boolean. We do three separate unit tests for
        versions 204 through 206 because mock doesn't like us altering the
        return_value in a loop.
        """
        with patch("subprocess.Popen") as popen_mock:
            _expected = True
            _version = 206
            _output = f"systemd {_version}\n-SYSVINIT"
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: (_output, None),
                pid=lambda: 12345,
                retcode=0,
            )
            # Ensure that os.stat returns True. os.stat doesn't return a bool
            # normally, but the code is doing a simple truth check on the
            # return data, so it is sufficient enough to mock it as True for
            # these tests.
            with patch("os.stat", side_effect=_booted_effect):
                with patch("salt.utils.systemd.status", return_value=True):
                    # Test without context dict passed
                    self.assertEqual(_systemd.has_scope(), _expected)
                    # Test that context key is set when context dict is passed
                    context = {"salt.utils.systemd.status": True}
                    self.assertEqual(_systemd.has_scope(context), _expected)
                    self.assertEqual(
                        context,
                        {
                            "salt.utils.systemd.booted": True,
                            "salt.utils.systemd.version": _version,
                            "salt.utils.systemd.status": True,
                        },
                    )

    def test_has_scope_no_systemd(self):
        """
        Test the case where the system is not systemd-booted. We should not be
        performing a version check in these cases as there is no need.
        """
        with patch("os.stat", side_effect=_not_booted_effect):
            # Test without context dict passed
            self.assertFalse(_systemd.has_scope())
            # Test that context key is set when context dict is passed.
            # Because we are not systemd-booted, there should be no key in the
            # context dict for the version check, as we shouldn't have
            # performed this check.
            context = {}
            self.assertFalse(_systemd.has_scope(context))
            self.assertEqual(context, {"salt.utils.systemd.booted": False})

    def test_has_scope_version_parse_problem(self):
        """
        Test the case where the system is systemd-booted, but we failed to
        parse the "systemctl --version" output.
        """
        with patch("subprocess.Popen") as popen_mock:
            popen_mock.return_value = Mock(
                communicate=lambda *args, **kwargs: ("invalid", None),
                pid=lambda: 12345,
                retcode=0,
            )
            with patch("os.stat", side_effect=_booted_effect):
                # Test without context dict passed
                self.assertFalse(_systemd.has_scope())
                # Test that context key is set when context dict is passed. A
                # failure to parse the systemctl output should not set a context
                # key, so it should not be present in the context dict.
                context = {}
                self.assertFalse(_systemd.has_scope(context))
                self.assertEqual(context, {"salt.utils.systemd.booted": True})

    def test_has_scope_invalid_context(self):
        """
        Test with invalid context data. The context value must be a dict, so
        this should raise a SaltInvocationError.
        """
        # Test with invalid context data
        with self.assertRaises(SaltInvocationError):
            _systemd.has_scope(99999)

    @patch("salt.utils.systemd.dbus", False)
    def test_pid_to_service_systemctl_valid_pid(self):
        """
        Test parsing of systemctl status output for a valid PID.
        """
        systemctl_output = """
        ● firewalld.service - firewalld - dynamic firewall daemon
        Loaded: loaded (/usr/lib/systemd/system/firewalld.service; enabled; vendor preset: disabled)
        {"_CAP_EFFECTIVE":"ffffffffff","_SYSTEMD_UNIT":"firewalld.service"}
        """
        subprocess_mock = Mock(return_value=Mock(stdout=systemctl_output))
        with patch("salt.utils.systemd.subprocess.run", subprocess_mock):
            assert _systemd.pid_to_service(1374) == "firewalld"

    @patch("salt.utils.systemd.dbus", False)
    def test_pid_to_service_systemctl_invalid_pid(self):
        """
        Test parsing of systemctl status output for an invalid PID.
        """
        subprocess_mock = Mock(
            side_effect=subprocess.CalledProcessError(
                returncode=1, cmd=["systemctl", "--output", "json", "status", "999999"]
            )
        )
        with patch("salt.utils.systemd.subprocess.run", subprocess_mock):
            # breakpoint()
            assert _systemd.pid_to_service(999999) is None

    def test_pid_to_service_dbus_valid_pid(self):
        """
        Test translating a valid PID to a service name via DBUS.
        """
        dbus_mock = Mock()
        dbus_interface_get_mock = Mock(return_value="firewalld.service")
        dbus_mock.Interface().Get = dbus_interface_get_mock
        with patch("salt.utils.systemd.dbus", dbus_mock):
            assert _systemd.pid_to_service(1374) == "firewalld"
            dbus_interface_get_mock.assert_called_with(
                "org.freedesktop.systemd1.Unit", "Id"
            )

    def test_pid_to_service_dbus_invalid_pid(self):
        """
        Test translating an invalid PID to a service name via DBUS.
        """

        class DBusException(Exception):
            """
            Raised by DBUS, e.g. when a PID does not belong to a service
            """

        dbus_mock = Mock()
        dbus_mock.DBusException = DBusException()
        dbus_mock.GetUnitByPID = Mock(site_effect=dbus_mock.DBusException)
        with patch("salt.utils.systemd.dbus", dbus_mock):
            assert _systemd.pid_to_service(99999) is None



import shutil

import pytest

from tests.conftest import FIPS_TESTRUN


@pytest.fixture(scope="package")
def pillar_state_tree(tmp_path_factory):
    _pillar_state_tree = tmp_path_factory.mktemp("pillar")
    try:
        yield _pillar_state_tree
    finally:
        shutil.rmtree(str(_pillar_state_tree), ignore_errors=True)


@pytest.fixture(scope="package")
def pillar_salt_master(salt_factories, pillar_state_tree):
    config_defaults = {
        "pillar_roots": {"base": [str(pillar_state_tree)]},
        "open_mode": True,
        "pillar_cache": True,
        "ext_pillar": [
            {"extra_minion_data_in_pillar": "*"},
        ],
    }
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "publish_signing_algorithm": (
            "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1"
        ),
    }
    factory = salt_factories.salt_master_daemon(
        "pillar-cache-functional-master",
        defaults=config_defaults,
        overrides=config_overrides,
    )
    with factory.started():
        yield factory


@pytest.fixture(scope="package")
def pillar_salt_minion(pillar_salt_master):
    assert pillar_salt_master.is_running()
    config_overrides = {
        "fips_mode": FIPS_TESTRUN,
        "encryption_algorithm": "OAEP-SHA224" if FIPS_TESTRUN else "OAEP-SHA1",
        "signing_algorithm": "PKCS1v15-SHA224" if FIPS_TESTRUN else "PKCS1v15-SHA1",
    }
    factory = pillar_salt_master.salt_minion_daemon(
        "pillar-cache-functional-minion-1",
        defaults={"open_mode": True, "hi": "there", "pass_to_ext_pillars": ["hi"]},
        overrides=config_overrides,
    )
    with factory.started():
        # Sync All
        salt_call_cli = factory.salt_call_cli()
        ret = salt_call_cli.run("saltutil.sync_all", _timeout=120)
        assert ret.returncode == 0, ret
        yield factory


@pytest.fixture(scope="package")
def pillar_salt_call_cli(pillar_salt_minion):
    return pillar_salt_minion.salt_call_cli()



"""
Tests for payload
"""

import pytest


@pytest.mark.slow_test
@pytest.mark.skip_if_not_root
@pytest.mark.skip_on_windows
@pytest.mark.skip_on_darwin
def test_payload_no_exception(salt_cli, salt_master, salt_minion):
    """
    Test to confirm that no exception is thrown with the jinja file
    when executed on the minion
    """
    test_set_hostname = """
    {%- set host = pillar.get("hostname", "UNKNOWN") %}
    {%- if host == 'UNKNOWN' %}
      {{ raise("Unsupported UNKNOWN hostname") }}
    {%- else %}
        hostnamectl set-hostname {{ host }}
    {%- endif %}
    """
    with salt_master.state_tree.base.temp_file("set_hostname.j2", test_set_hostname):

        ret = salt_cli.run("test.ping", minion_tgt=salt_minion.id)
        assert ret.returncode == 0
        assert ret.data is True

        ret = salt_cli.run(
            "cmd.script",
            "salt://set_hostname.j2",
            "template=jinja",
            pillar={"hostname": "test"},
            minion_tgt=salt_minion.id,
        )
        assert "AttributeError:" not in ret.stdout



"""
Test Salt MySQL state module across various MySQL variants
"""

import logging
import time

import pytest
from pytestshellutils.utils import format_callback_to_string
from saltfactories.utils.functional import StateResult

import salt.modules.mysql as mysqlmod
from tests.support.pytest.mysql import *  # pylint: disable=wildcard-import,unused-wildcard-import

log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("dockerd"),
    pytest.mark.skipif(
        mysqlmod.MySQLdb is None, reason="No python mysql client installed."
    ),
]


def _get_mysql_error(context):
    return context.pop("mysql.error", None)


class CallWrapper:
    def __init__(self, func, container, ctx):
        self.func = func
        self.container = container
        self.ctx = ctx

    def __call__(self, *args, **kwargs):
        kwargs.update(self.container.get_credentials(**kwargs))
        retry = 1
        retries = 3
        ret = None
        while True:
            ret = self.func(*list(args), **kwargs.copy())
            mysql_error = _get_mysql_error(self.ctx)
            if mysql_error is None:
                break

            retry += 1

            if retry > retries:
                break

            time.sleep(0.5)
            log.debug(
                "Retrying(%s out of %s) %s because of the following error: %s",
                retry,
                retries,
                format_callback_to_string(self.func, args, kwargs),
                mysql_error,
            )
        return ret


class StateSingleWrapper:
    def __init__(self, func, container, ctx):
        self.func = func
        self.container = container
        self.ctx = ctx

    def __call__(self, statefunc, *args, **kwargs):
        if statefunc.startswith("mysql_"):
            kwargs.update(self.container.get_credentials(**kwargs))
            retry = 1
            retries = 3
            ret = None
            while True:
                ret = self.func(statefunc, *list(args), **kwargs.copy())
                if ret.result:
                    break

                mysql_error = _get_mysql_error(self.ctx)
                if mysql_error is None:
                    break

                retry += 1
                if retry > retries:
                    break

                time.sleep(0.5)
                log.debug(
                    "Retrying(%s out of %s) %s because of the following return: %s",
                    retry,
                    retries,
                    format_callback_to_string(statefunc, args, kwargs),
                    ret,
                )
        else:
            # No retries for every other state function
            ret = self.func(statefunc, *args, **kwargs)
        if isinstance(ret, StateResult):
            # Sadly, because we're wrapping, we need to return the raw
            # attribute for a StateResult class to be recreated.
            return ret.raw
        return ret


@pytest.fixture(scope="module")
def mysql(modules, mysql_container, loaders):
    for name in list(modules):
        if name.startswith("mysql."):
            modules._dict[name] = CallWrapper(
                modules._dict[name],
                mysql_container,
                loaders.context,
            )
        if name == "state.single":
            modules._dict[name] = StateSingleWrapper(
                modules._dict[name],
                mysql_container,
                loaders.context,
            )
    return modules.mysql


@pytest.fixture(scope="module")
def mysql_states(mysql, states, mysql_container):
    # Just so we also have the container running
    return states


@pytest.fixture(scope="module")
def mysql_user(mysql_states):
    return mysql_states.mysql_user


@pytest.fixture(scope="module")
def mysql_query(mysql_states):
    return mysql_states.mysql_query


@pytest.fixture(scope="module")
def mysql_grants(mysql_states):
    return mysql_states.mysql_grants


@pytest.fixture(scope="module")
def mysql_database(mysql_states):
    return mysql_states.mysql_database


def test_database_present_absent(mysql_database):
    ret = mysql_database.present(name="test_database")
    assert ret.changes
    assert ret.changes == {"test_database": "Present"}
    assert ret.comment
    assert ret.comment == "The database test_database has been created"

    ret = mysql_database.absent(name="test_database")
    assert ret.changes
    assert ret.changes == {"test_database": "Absent"}
    assert ret.comment
    assert ret.comment == "Database test_database has been removed"


def test_grants_present_absent(mysql, mysql_grants):

    # Create the database
    ret = mysql.db_create("salt")
    assert ret

    # Create a user
    ret = mysql.user_create(
        "george",
        host="localhost",
        password="badpassword",
    )
    assert ret

    try:

        ret = mysql_grants.present(
            name="add_salt_grants",
            grant="select,insert,update",
            database="salt.*",
            user="george",
            host="localhost",
        )
        assert ret.changes
        assert ret.changes == {"add_salt_grants": "Present"}
        assert ret.comment
        assert (
            ret.comment
            == "Grant select,insert,update on salt.* to george@localhost has been added"
        )

        ret = mysql_grants.absent(
            name="delete_salt_grants",
            grant="select,insert,update",
            database="salt.*",
            user="george",
            host="localhost",
        )
        assert ret.changes
        assert ret.changes == {"delete_salt_grants": "Absent"}
        assert ret.comment
        assert (
            ret.comment
            == "Grant select,insert,update on salt.* for george@localhost has been revoked"
        )

    finally:
        # Remove the user
        ret = mysql.user_remove("george", host="localhost")
        assert ret

        # Remove the database
        ret = mysql.db_remove("salt")
        assert ret


def test_grants_present_absent_all_privileges(mysql, mysql_grants):

    # Create the database
    ret = mysql.db_create("salt_2_0")
    assert ret

    # Create a user
    ret = mysql.user_create(
        "george",
        host="localhost",
        password="badpassword",
    )
    assert ret

    try:
        ret = mysql_grants.present(
            name="add_salt_grants",
            grant="all privileges",
            database="salt_2_0.*",
            user="george",
            host="localhost",
        )

        assert ret.changes
        assert ret.changes == {"add_salt_grants": "Present"}

        assert ret.comment
        assert (
            ret.comment
            == "Grant all privileges on salt_2_0.* to george@localhost has been added"
        )

        ret = mysql_grants.absent(
            name="delete_salt_grants",
            grant="all privileges",
            database="salt_2_0.*",
            user="george",
            host="localhost",
        )

        assert ret.changes
        assert ret.changes == {"delete_salt_grants": "Absent"}

        assert ret.comment
        assert (
            ret.comment
            == "Grant all privileges on salt_2_0.* for george@localhost has been revoked"
        )

    finally:
        # Remove the user
        ret = mysql.user_remove("george", host="localhost")
        assert ret

        # Remove the database
        ret = mysql.db_remove("salt_2_0")
        assert ret


def test_user_present_absent(mysql_user):

    ret = mysql_user.present(
        name="george",
        host="localhost",
        password="password",
    )
    assert ret.changes
    assert ret.changes == {"george": "Present"}
    assert ret.comment
    assert ret.comment == "The user george@localhost has been added"

    ret = mysql_user.absent(
        name="george",
        host="localhost",
    )
    assert ret.changes
    assert ret.changes == {"george": "Absent"}
    assert ret.comment
    assert ret.comment == "User george@localhost has been removed"


def test_user_present_absent_passwordless(mysql_user):

    ret = mysql_user.present(
        name="george",
        host="localhost",
        allow_passwordless=True,
    )
    assert ret.changes
    assert ret.changes == {"george": "Present"}
    assert ret.comment
    assert (
        ret.comment
        == "The user george@localhost has been added with passwordless login"
    )

    ret = mysql_user.absent(
        name="george",
        host="localhost",
    )
    assert ret.changes
    assert ret.changes == {"george": "Absent"}
    assert ret.comment
    assert ret.comment == "User george@localhost has been removed"


def test_user_present_absent_unixsocket(mysql, mysql_user, mysql_container):

    # The auth_socket plugin on MariaDB is unavailable
    # on versions 10.1 - 10.3
    if "mariadb" in mysql_container.mysql_name:
        if mysql_container.mysql_version in ("10.1", "10.2", "10.3"):
            pytest.skip(
                "The auth_socket plugin is unavaiable "
                "for the {}:{} docker image.".format(
                    mysql_container.mysql_name, mysql_container.mysql_version
                )
            )

    # enable the auth_socket plugin on MySQL
    # already enabled on MariaDB > 10.3
    try:
        if "mariadb" not in mysql_container.mysql_name:
            ret = mysql.plugin_add("auth_socket")
            assert ret

        ret = mysql_user.present(
            name="george",
            host="localhost",
            unix_socket=True,
            allow_passwordless=False,
        )
        assert ret.changes
        assert ret.changes == {"george": "Present"}
        assert ret.comment
        assert (
            ret.comment == "The user george@localhost has been added using unix_socket"
        )

        ret = mysql_user.absent(
            name="george",
            host="localhost",
        )
        assert ret.changes
        assert ret.changes == {"george": "Absent"}
        assert ret.comment
        assert ret.comment == "User george@localhost has been removed"
    finally:
        if "mariadb" not in mysql_container.mysql_name:
            ret = mysql.plugin_remove("auth_socket")
            assert ret


def test_grants_present_absent_state_file(
    modules, mysql, mysql_grants, mysql_combo, state_tree
):

    content = """
    sbclient_2_0:
      mysql_user.present:
        - host: localhost
        - password: sbclient
        - connection_user: {mysql_user}
        - connection_pass: {mysql_pass}
        - connection_db: mysql
        - connection_port: {mysql_port}
      mysql_database.present:
        - connection_user: {mysql_user}
        - connection_pass: {mysql_pass}
        - connection_db: mysql
        - connection_port: {mysql_port}
      mysql_grants.present:
        - grant: ALL PRIVILEGES
        - user: sbclient_2_0
        - database: sbclient_2_0.*
        - host: localhost
        - connection_user: {mysql_user}
        - connection_pass: {mysql_pass}
        - connection_db: mysql
        - connection_port: {mysql_port}
    """.format(
        mysql_user=mysql_combo.mysql_root_user,
        mysql_pass=mysql_combo.mysql_root_passwd,
        mysql_port=mysql_combo.mysql_port,
    )

    try:
        with pytest.helpers.temp_file("manage_mysql.sls", content, state_tree):
            ret = modules.state.apply("manage_mysql")

            # Check user creation
            state = "mysql_user_|-sbclient_2_0_|-sbclient_2_0_|-present"
            assert state in ret
            assert "changes" in ret[state]
            assert ret[state].changes == {"sbclient_2_0": "Present"}

            # Check database creation
            state = "mysql_database_|-sbclient_2_0_|-sbclient_2_0_|-present"
            assert state in ret
            assert "changes" in ret[state]
            assert ret[state].changes == {"sbclient_2_0": "Present"}

            # Check grant creation
            state = "mysql_grants_|-sbclient_2_0_|-sbclient_2_0_|-present"
            assert state in ret
            assert "comment" in ret[state]
            assert (
                ret[state].comment
                == "Grant ALL PRIVILEGES on sbclient_2_0.* to sbclient_2_0@localhost has been added"
            )

        ret = mysql_grants.absent(
            name="delete_sbclient_grants",
            grant="all privileges",
            database="sbclient_2_0.*",
            user="sbclient_2_0",
            host="localhost",
        )
        assert ret.changes
        assert ret.changes == {"delete_sbclient_grants": "Absent"}

        assert ret.comment
        assert (
            ret.comment
            == "Grant all privileges on sbclient_2_0.* for sbclient_2_0@localhost has been revoked"
        )

    finally:
        # Remove the user
        ret = mysql.user_remove("sbclient_2_0", host="localhost")
        assert ret

        # Remove the database
        ret = mysql.db_remove("sbclient_2_0")
        assert ret



"""
tests.pytests.functional.cli.test_batch
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import salt.cli.batch
import salt.config
import salt.utils.jid
from tests.support.mock import Mock, patch


class MockPub:
    """
    Mock salt.client.LocalClient.pub method
    """

    calls = 0
    initial_ping = False
    batch1_jid = None
    batch1_tgt = None
    batch2_jid = None
    batch2_tgt = None
    batch3_jid = None
    batch3_tgt = None

    def __call__(self, tgt, fun, *args, **kwargs):
        if tgt == "minion*" and fun == "test.ping":
            MockPub.calls += 1
            MockPub.initial_ping = salt.utils.jid.gen_jid({})
            pub_ret = {
                "jid": MockPub.initial_ping,
                "minions": ["minion0", "minion1", "minion2", "minion3"],
            }
        elif fun == "state.sls":
            if MockPub.calls == 1:
                MockPub.calls += 1
                MockPub.batch1_tgt = list(tgt)
                MockPub.batch1_jid = jid = salt.utils.jid.gen_jid({})
                pub_ret = {"jid": jid, "minions": tgt}
            elif MockPub.calls == 2:
                MockPub.calls += 1
                MockPub.batch2_tgt = tgt
                MockPub.batch2_jid = jid = salt.utils.jid.gen_jid({})
                pub_ret = {"jid": jid, "minions": tgt}
            elif MockPub.calls == 3:
                MockPub.calls += 1
                MockPub.batch3_tgt = tgt
                MockPub.batch3_jid = jid = salt.utils.jid.gen_jid({})
                pub_ret = {"jid": jid, "minions": tgt}
        elif fun == "saltutil.find_job":
            jid = salt.utils.jid.gen_jid({})
            pub_ret = {"jid": jid, "minions": tgt}
        return pub_ret


class MockSubscriber:
    """
    Mock salt.transport.ipc IPCMessageSubscriber in order to inject events into
    salt.utils.Event
    """

    calls = 0
    pubret = None

    def __init__(self, *args, **kwargs):
        return

    def read(self, timeout=None):
        """
        Mock IPCMessageSubcriber read method.

        - Return events for initial ping
        - Returns event for a minion in first batch to cause second batch to get sent.
        - Returns 5 null events on first iteration of second batch to go back to first batch.
        - On second iteration of first batch, send an event from second batch which will get cached.
        - Return events for the rest of the batches.
        """
        if MockSubscriber.pubret.initial_ping:
            # Send ping responses for 4 minions
            jid = MockSubscriber.pubret.initial_ping
            if MockSubscriber.calls == 0:
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id="minion0", fun="test.ping")
            elif MockSubscriber.calls == 1:
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id="minion1", fun="test.ping")
            elif MockSubscriber.calls == 2:
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id="minion2", fun="test.ping")
            elif MockSubscriber.calls == 3:
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id="minion3", fun="test.ping")
        if MockSubscriber.pubret.batch1_jid:
            jid = MockSubscriber.pubret.batch1_jid
            tgt = MockSubscriber.pubret.batch1_tgt
            if MockSubscriber.calls == 4:
                # Send a return for first minion in first batch. This causes the
                # second batch to get sent.
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id=tgt[0], fun="state.sls")
        if MockSubscriber.pubret.batch2_jid:
            if MockSubscriber.calls <= 10:
                # Skip the first iteration of the second batch; this will cause
                # batch logic to go back to iterating over the first batch.
                MockSubscriber.calls += 1
                return
            elif MockSubscriber.calls == 11:
                # Send the minion from the second batch, This event will get cached.
                jid = MockSubscriber.pubret.batch2_jid
                tgt = MockSubscriber.pubret.batch2_tgt
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id=tgt[0], fun="state.sls")
        if MockSubscriber.calls == 12:
            jid = MockSubscriber.pubret.batch1_jid
            tgt = MockSubscriber.pubret.batch1_tgt
            MockSubscriber.calls += 1
            return self._ret(jid, minion_id=tgt[1], fun="state.sls")
        if MockSubscriber.pubret.batch3_jid:
            jid = MockSubscriber.pubret.batch3_jid
            tgt = MockSubscriber.pubret.batch3_tgt
            if MockSubscriber.calls == 13:
                MockSubscriber.calls += 1
                return self._ret(jid, minion_id=tgt[0], fun="state.sls")
        return

    def _ret(self, jid, minion_id, fun, _return=True, _retcode=0):
        """
        Create a mock return from a jid, minion, and fun
        """
        dumped = salt.payload.dumps(
            {
                "fun_args": [],
                "jid": jid,
                "return": _return,
                "retcode": 0,
                "success": True,
                "cmd": "_return",
                "fun": fun,
                "id": minion_id,
                "_stamp": "2021-05-24T01:23:25.373194",
            },
            use_bin_type=True,
        )
        tag = f"salt/job/{jid}/ret/{minion_id}".encode()
        return b"".join([tag, b"\n\n", dumped])

    def connect(self, timeout=None):
        pass


def test_batch_issue_56273():
    """
    Regression test for race condition in batch logic.
    https://github.com/saltstack/salt/issues/56273
    """

    mock_pub = MockPub()
    MockSubscriber.pubret = mock_pub

    def returns_for_job(jid):
        return True

    opts = {
        "conf_file": "",
        "tgt": "minion*",
        "fun": "state.sls",
        "arg": ["foo"],
        "timeout": 1,
        "gather_job_timeout": 1,
        "batch": 2,
        "extension_modules": "",
        "failhard": True,
    }
    with patch("salt.transport.ipc.IPCMessageSubscriber", MockSubscriber):
        batch = salt.cli.batch.Batch(opts, quiet=True)
        with patch.object(batch.local, "pub", Mock(side_effect=mock_pub)):
            with patch.object(
                batch.local, "returns_for_job", Mock(side_effect=returns_for_job)
            ):
                ret = list(batch.run())
    assert len(ret) == 4
    for val, _ in ret:
        values = list(val.values())
        assert len(values) == 1
        assert values[0] is True



"""
Management of OpenStack Keystone Domains
========================================

.. versionadded:: 2018.3.0

:depends: shade
:configuration: see :py:mod:`salt.modules.keystoneng` for setup instructions

Example States

.. code-block:: yaml

    create domain:
      keystone_domain.present:
        - name: domain1

    create domain with optional params:
      keystone_domain.present:
        - name: domain1
        - enabled: False
        - description: 'my domain'

    delete domain:
      keystone_domain.absent:
        - name: domain1
"""

__virtualname__ = "keystone_domain"


def __virtual__():
    if "keystoneng.domain_get" in __salt__:
        return __virtualname__
    return (
        False,
        "The keystoneng execution module failed to load: shade python module is not"
        " available",
    )


def present(name, auth=None, **kwargs):
    """
    Ensure domain exists and is up-to-date

    name
        Name of the domain

    enabled
        Boolean to control if domain is enabled

    description
        An arbitrary description of the domain
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    kwargs = __utils__["args.clean_kwargs"](**kwargs)

    __salt__["keystoneng.setup_clouds"](auth)

    domain = __salt__["keystoneng.domain_get"](name=name)

    if not domain:
        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = kwargs
            ret["comment"] = f"Domain {name} will be created."
            return ret

        kwargs["name"] = name
        domain = __salt__["keystoneng.domain_create"](**kwargs)
        ret["changes"] = domain
        ret["comment"] = "Created domain"
        return ret

    changes = __salt__["keystoneng.compare_changes"](domain, **kwargs)
    if changes:
        if __opts__["test"]:
            ret["result"] = None
            ret["changes"] = changes
            ret["comment"] = f"Domain {name} will be updated."
            return ret

        kwargs["domain_id"] = domain.id
        __salt__["keystoneng.domain_update"](**kwargs)
        ret["changes"].update(changes)
        ret["comment"] = "Updated domain"

    return ret


def absent(name, auth=None):
    """
    Ensure domain does not exist

    name
        Name of the domain
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    __salt__["keystoneng.setup_clouds"](auth)

    domain = __salt__["keystoneng.domain_get"](name=name)

    if domain:
        if __opts__["test"] is True:
            ret["result"] = None
            ret["changes"] = {"name": name}
            ret["comment"] = f"Domain {name} will be deleted."
            return ret

        __salt__["keystoneng.domain_delete"](name=domain)
        ret["changes"]["id"] = domain.id
        ret["comment"] = "Deleted domain"

    return ret



import os

import pytest

pytestmark = [
    pytest.mark.skip_unless_on_windows,
]


def test_ssm_present(install_salt):
    """
    The ssm.exe binary needs to be present in both the zip and the exe/msi
    builds
    """
    assert os.path.exists(install_salt.ssm_bin)



# pylint: disable=W1401
'''
This module renders highstate configuration into a more human readable format.

How it works:

`highstate or lowstate` data is parsed with a `processor` this defaults to `highstate_doc.processor_markdown`.
The processed data is passed to a `jinja` template that builds up the document content.


configuration: Pillar

.. code-block:: none

    # the following defaults can be overridden
    highstate_doc.config:

        # list of regex of state names to ignore in `highstate_doc.process_lowstates`
        filter_id_regex:
            - '.*!doc_skip$'

        # list of regex of state functions to ignore in `highstate_doc.process_lowstates`
        filter_state_function_regex:
            - 'file.accumulated'

        # dict of regex to replace text after `highstate_doc.render`. (remove passwords)
        text_replace_regex:
            'password:.*^': '[PASSWORD]'

        # limit size of files that can be included in doc (10000 bytes)
        max_render_file_size: 10000

        # advanced option to set a custom lowstate processor
        processor: highstate_doc.processor_markdown


State example

.. code-block:: yaml

    {{sls}} note:
        highstate_doc.note:
            - name: example
            - order: 0
            - contents: |
                example `highstate_doc.note`
                ------------------
                This state does not do anything to the system! It is only used by a `processor`
                you can use `requisites` and `order` to move your docs around the rendered file.

    {{sls}} a file we don't want in the doc !doc_skip:
        file.managed:
            - name: /root/passwords
            - contents: 'password: sadefgq34y45h56q'
            # also could use `highstate_doc.config: text_replace_regex` to replace
            # password string. `password:.*^': '[PASSWORD]`


To create the help document build a State that uses `highstate_doc.render`.
For performance it's advised to not included this state in your `top.sls` file.

.. code-block:: yaml

    # example `salt://makereadme.sls`
    make helpfile:
        file.managed:
            - name: /root/README.md
            - contents: {{salt.highstate_doc.render()|json}}
            - show_diff: {{opts['test']}}
            - mode: '0640'
            - order: last

Run our `makereadme.sls` state to create `/root/README.md`.

.. code-block:: bash

    # first ensure `highstate` return without errors or changes
    salt-call state.highstate
    salt-call state.apply makereadme
    # or if you don't want the extra `make helpfile` state
    salt-call --out=newline_values_only salt.highstate_doc.render > /root/README.md ; chmod 0600 /root/README.md


Creating a document collection
------------------------------

From the master we can run the following script to
creates a collection of all your minion documents.

.. code-block:: bash

    salt '*' state.apply makereadme

.. code-block:: python

    #!/bin/python
    import os
    import salt.client
    s = salt.client.LocalClient()
    # NOTE: because of issues with `cp.push` use `highstate_doc.read_file`
    o = s.cmd('*', 'highstate_doc.read_file', ['/root/README.md'])
    for m in o:
        d = o.get(m)
        if d and not d.endswith('is not available.'):
            # mkdir m
            #directory = os.path.dirname(file_path)
            if not os.path.exists(m):
                os.makedirs(m)
            with open(m + '/README.md','wb') as f:
                f.write(d)
            print('ADDED: ' + m + '/README.md')


Once the master has a collection of all the README files.
You can use pandoc to create HTML versions of the markdown.

.. code-block:: bash

    # process all the readme.md files to readme.html
    if which pandoc; then echo "Found pandoc"; else echo "** Missing pandoc"; exit 1; fi
    if which gs; then echo "Found gs"; else echo "** Missing gs(ghostscript)"; exit 1; fi
    readme_files=$(find $dest -type f -path "*/README.md" -print)
    for f in $readme_files ; do
        ff=${f#$dest/}
        minion=${ff%%/*}
        echo "process: $dest/${minion}/$(basename $f)"
        cat $dest/${minion}/$(basename $f) | \
            pandoc --standalone --from markdown_github --to html \
            --include-in-header $dest/style.html \
            > $dest/${minion}/$(basename $f).html
    done

It is also nice to put the help files in source control.

    # git init
    git add -A
    git commit -am 'updated docs'
    git push -f


Other hints
-----------

If you wish to customize the document format:

.. code-block:: none

    # you could also create a new `processor` for perhaps reStructuredText
    # highstate_doc.config:
    #     processor: doc_custom.processor_rst

    # example `salt://makereadme.jinja`
    """
    {{opts['id']}}
    ==========================================

    {# lowstates is set from highstate_doc.render() #}
    {# if lowstates is missing use salt.highstate_doc.process_lowstates() #}
    {% for s in lowstates %}
    {{s.id}}
    -----------------------------------------------------------------
    {{s.function}}

    {{s.markdown.requisite}}
    {{s.markdown.details}}

    {%- endfor %}
    """

    # example `salt://makereadme.sls`
    {% import_text "makereadme.jinja" as makereadme %}
    {{sls}} or:
        file.managed:
            - name: /root/README_other.md
            - contents: {{salt.highstate_doc.render(jinja_template_text=makereadme)|json}}
            - mode: '0640'


Some `replace_text_regex` values that might be helpful::

    CERTS
    -----

    ``'-----BEGIN RSA PRIVATE KEY-----[\\r\\n\\t\\f\\S]{0,2200}': 'XXXXXXX'``
    ``'-----BEGIN CERTIFICATE-----[\\r\\n\\t\\f\\S]{0,2200}': 'XXXXXXX'``
    ``'-----BEGIN DH PARAMETERS-----[\\r\\n\\t\\f\\S]{0,2200}': 'XXXXXXX'``
    ``'-----BEGIN PRIVATE KEY-----[\\r\\n\\t\\f\\S]{0,2200}': 'XXXXXXX'``
    ``'-----BEGIN OPENSSH PRIVATE KEY-----[\\r\\n\\t\\f\\S]{0,2200}': 'XXXXXXX'``
    ``'ssh-rsa .* ': 'ssh-rsa XXXXXXX '``
    ``'ssh-dss .* ': 'ssh-dss XXXXXXX '``

    DB
    --

    ``'DB_PASS.*': 'DB_PASS = XXXXXXX'``
    ``'5432:*:*:.*': '5432:*:XXXXXXX'``
    ``"'PASSWORD': .*": "'PASSWORD': 'XXXXXXX',"``
    ``" PASSWORD '.*'": " PASSWORD 'XXXXXXX'"``
    ``'PGPASSWORD=.* ': 'PGPASSWORD=XXXXXXX'``
    ``"_replication password '.*'":  "_replication password 'XXXXXXX'"``

    OTHER
    -----

    ``'EMAIL_HOST_PASSWORD =.*': 'EMAIL_HOST_PASSWORD =XXXXXXX'``
    ``"net ads join -U '.*@MFCFADS.MATH.EXAMPLE.CA.* ": "net ads join -U '.*@MFCFADS.MATH.EXAMPLE.CA%XXXXXXX "``
    ``"net ads join -U '.*@NEXUS.EXAMPLE.CA.* ": "net ads join -U '.*@NEXUS.EXAMPLE.CA%XXXXXXX "``
    ``'install-uptrack .* --autoinstall': 'install-uptrack XXXXXXX --autoinstall'``
    ``'accesskey = .*': 'accesskey = XXXXXXX'``
    ``'auth_pass .*': 'auth_pass XXXXXXX'``
    ``'PSK "0x.*': 'PSK "0xXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'``
    ``'SECRET_KEY.*': 'SECRET_KEY = XXXXXXX'``
    ``"password=.*": "password=XXXXXXX"``
    ``'<password>.*</password>': '<password>XXXXXXX</password>'``
    ``'<salt>.*</salt>': '<salt>XXXXXXX</salt>'``
    ``'application.secret = ".*"': 'application.secret = "XXXXXXX"'``
    ``'url = "postgres://.*"': 'url = "postgres://XXXXXXX"'``
    ``'PASS_.*_PASS': 'PASS_XXXXXXX_PASS'``

    HTACCESS
    --------

    ``':{PLAIN}.*': ':{PLAIN}XXXXXXX'``

'''


import logging
import re

import salt.utils.files
import salt.utils.stringutils
import salt.utils.templates as tpl
import salt.utils.yaml

__virtualname__ = "highstate_doc"

log = logging.getLogger(__name__)


markdown_basic_jinja_template_txt = """
{% for s in lowstates %}
`{{s.id_full}}`
-----------------------------------------------------------------
 * state: {{s.state_function}}
 * name: `{{s.name}}`

{{s.markdown.requisites}}
{{s.markdown.details}}

{%- endfor %}
"""

markdown_default_jinja_template_txt = (
    """
Configuration Managment
===============================================================================

```
####################################################
fqdn: {{grains.get('fqdn')}}
os: {{grains.get('os')}}
osfinger: {{grains.get('osfinger')}}
mem_total: {{grains.get('mem_total')}}MB
num_cpus: {{grains.get('num_cpus')}}
ipv4: {{grains.get('ipv4')}}
master: {{opts.get('master')}}
####################################################
```

This system is fully or partly managed using Salt.

The following sections are a rendered view of what the configuration management system
controlled on this system. Each item is handled in order from top to bottom unless some
requisites like `require` force other ordering.

"""
    + markdown_basic_jinja_template_txt
)


markdown_advanced_jinja_template_txt = (
    markdown_default_jinja_template_txt
    + r"""

{% if vars.get('doc_other', True) -%}
Other information
=====================================================================================

```

salt grain: ip_interfaces
-----------------------------------------------------------------
{{grains['ip_interfaces']|dictsort}}


salt grain: hwaddr_interfaces
-----------------------------------------------------------------
{{grains['hwaddr_interfaces']|dictsort}}

{% if not grains['os'] == 'Windows' %}

{% if salt['cmd.has_exec']('ip') -%}
# ip address show
-----------------------------------------------------------------
{{salt['cmd.run']('ip address show | sed "/valid_lft/d"')}}


# ip route list table all
-----------------------------------------------------------------
{{salt['cmd.run']('ip route list table all')}}
{% endif %}

{% if salt['cmd.has_exec']('iptables') %}
{%- if salt['cmd.has_exec']('iptables-save') -%}
# iptables-save
-----------------------------------------------------------------
{{salt['cmd.run']("iptables --list > /dev/null; iptables-save | \grep -v -F '#' | sed '/^:/s@\[[0-9]\{1,\}:[0-9]\{1,\}\]@[0:0]@g'")}}


# ip6tables-save
-----------------------------------------------------------------
{{salt['cmd.run']("ip6tables --list > /dev/null; ip6tables-save | \grep -v -F '#' | sed '/^:/s@\[[0-9]\{1,\}:[0-9]\{1,\}\]@[0:0]@g'")}}
{%- else -%}
# iptables --list-rules
-----------------------------------------------------------------
{{salt['cmd.run']('iptables --list-rules')}}


# ip6tables --list-rules
-----------------------------------------------------------------
{{salt['cmd.run']('ip6tables --list-rules')}}
{% endif %}
{% endif %}

{% if salt['cmd.has_exec']('firewall-cmd') -%}
# firewall-cmd --list-all
-----------------------------------------------------------------
{{salt['cmd.run']('firewall-cmd --list-all')}}
{% endif %}

# mount
-----------------------------------------------------------------
{{salt['cmd.run']('mount')}}

{% endif %}
"""
)


def markdown_basic_jinja_template(**kwargs):
    """
    Return text for a simple markdown jinja template

    This function can be used from the `highstate_doc.render` modules `jinja_template_function` option.
    """
    return markdown_basic_jinja_template_txt


def markdown_default_jinja_template(**kwargs):
    """
    Return text for a markdown jinja template that included a header

    This function can be used from the `highstate_doc.render` modules `jinja_template_function` option.
    """
    return markdown_default_jinja_template_txt


def markdown_full_jinja_template(**kwargs):
    """
    Return text for an advanced markdown jinja template

    This function can be used from the `highstate_doc.render` modules `jinja_template_function` option.
    """
    return markdown_advanced_jinja_template_txt


def _get_config(**kwargs):
    """
    Return configuration
    """
    config = {
        "filter_id_regex": [".*!doc_skip"],
        "filter_function_regex": [],
        "replace_text_regex": {},
        "processor": "highstate_doc.processor_markdown",
        "max_render_file_size": 10000,
        "note": None,
    }
    if "__salt__" in globals():
        config_key = f"{__virtualname__}.config"
        config.update(__salt__["config.get"](config_key, {}))
    # pylint: disable=C0201
    for k in set(config.keys()) & set(kwargs.keys()):
        config[k] = kwargs[k]
    return config


def read_file(name):
    """
    output the contents of a file:

    this is a workaround if the cp.push module does not work.
    https://github.com/saltstack/salt/issues/37133

    help the master output the contents of a document
    that might be saved on the minions filesystem.

    .. code-block:: python

        #!/bin/python
        import os
        import salt.client
        s = salt.client.LocalClient()
        o = s.cmd('*', 'highstate_doc.read_file', ['/root/README.md'])
        for m in o:
            d = o.get(m)
            if d and not d.endswith('is not available.'):
                # mkdir m
                #directory = os.path.dirname(file_path)
                if not os.path.exists(m):
                    os.makedirs(m)
                with open(m + '/README.md','wb') as fin:
                    fin.write(d)
                print('ADDED: ' + m + '/README.md')
    """
    out = ""
    try:
        with salt.utils.files.fopen(name, "r") as f:
            out = salt.utils.stringutils.to_unicode(f.read())
    except Exception as ex:  # pylint: disable=broad-except
        log.error(ex)
        return None
    return out


def render(
    jinja_template_text=None,
    jinja_template_function="highstate_doc.markdown_default_jinja_template",
    **kwargs,
):
    """
    Render highstate to a text format (default Markdown)

    if `jinja_template_text` is not set, `jinja_template_function` is used.

    jinja_template_text: jinja text that the render uses to create the document.
    jinja_template_function: a salt module call that returns template text.

    :options:
        highstate_doc.markdown_basic_jinja_template
        highstate_doc.markdown_default_jinja_template
        highstate_doc.markdown_full_jinja_template

    """
    config = _get_config(**kwargs)
    lowstates = process_lowstates(**kwargs)
    # TODO: __env__,
    context = {
        "saltenv": None,
        "config": config,
        "lowstates": lowstates,
        "salt": __salt__,
        "pillar": __pillar__,
        "grains": __grains__,
        "opts": __opts__,
        "kwargs": kwargs,
    }
    template_text = jinja_template_text
    if template_text is None and jinja_template_function:
        template_text = __salt__[jinja_template_function](**kwargs)
    if template_text is None:
        raise Exception("No jinja template text")

    txt = tpl.render_jinja_tmpl(template_text, context, tmplpath=None)
    # after processing the template replace passwords or other data.
    rt = config.get("replace_text_regex")
    for r in rt:
        txt = re.sub(r, rt[r], txt)
    return txt


def _blacklist_filter(s, config):
    ss = s["state"]
    sf = s["fun"]
    state_function = "{}.{}".format(s["state"], s["fun"])
    for b in config["filter_function_regex"]:
        if re.match(b, state_function):
            return True
    for b in config["filter_id_regex"]:
        if re.match(b, s["__id__"]):
            return True
    return False


def process_lowstates(**kwargs):
    """
    return processed lowstate data that was not blacklisted

    render_module_function is used to provide your own.
    defaults to from_lowstate
    """
    states = []
    config = _get_config(**kwargs)
    processor = config.get("processor")
    ls = __salt__["state.show_lowstate"]()

    if not isinstance(ls, list):
        raise Exception(
            "ERROR: to see details run: [salt-call state.show_lowstate]"
            " <-----***-SEE-***"
        )
    else:
        if ls:
            if not isinstance(ls[0], dict):
                raise Exception(
                    "ERROR: to see details run: [salt-call state.show_lowstate]"
                    " <-----***-SEE-***"
                )

    for s in ls:
        if _blacklist_filter(s, config):
            continue
        doc = __salt__[processor](s, config, **kwargs)
        states.append(doc)
    return states


def _state_data_to_yaml_string(data, whitelist=None, blacklist=None):
    """
    return a data dict in yaml string format.
    """
    y = {}
    if blacklist is None:
        # TODO: use salt defined STATE_REQUISITE_IN_KEYWORDS  STATE_RUNTIME_KEYWORDS  STATE_INTERNAL_KEYWORDS
        blacklist = [
            "__env__",
            "__id__",
            "__sls__",
            "fun",
            "name",
            "context",
            "order",
            "state",
            "require",
            "require_in",
            "watch",
            "watch_in",
        ]
    kset = set(data.keys())
    if blacklist:
        kset -= set(blacklist)
    if whitelist:
        kset &= set(whitelist)
    for k in kset:
        y[k] = data[k]
    if not y:
        return None
    return salt.utils.yaml.safe_dump(y, default_flow_style=False)


def _md_fix(text):
    """
    sanitize text data that is to be displayed in a markdown code block
    """
    return text.replace("```", "``[`][markdown parse fix]")


def _format_markdown_system_file(filename, config):
    ret = ""
    file_stats = __salt__["file.stats"](filename)
    y = _state_data_to_yaml_string(
        file_stats, whitelist=["user", "group", "mode", "uid", "gid", "size"]
    )
    if y:
        ret += f"file stat {filename}\n```\n{y}```\n"
    file_size = file_stats.get("size")
    if file_size <= config.get("max_render_file_size"):
        is_binary = True
        try:
            # TODO: this is linux only should find somthing portable
            file_type = __salt__["cmd.shell"](f"\\file -i '{filename}'")
            if "charset=binary" not in file_type:
                is_binary = False
        except Exception as ex:  # pylint: disable=broad-except
            # likely on a windows system, set as not binary for now.
            is_binary = False
        if is_binary:
            file_data = "[[skipped binary data]]"
        else:
            with salt.utils.files.fopen(filename, "r") as f:
                file_data = salt.utils.stringutils.to_unicode(f.read())
        file_data = _md_fix(file_data)
        ret += f"file data {filename}\n```\n{file_data}\n```\n"
    else:
        ret += "```\n{}\n```\n".format(
            "SKIPPED LARGE FILE!\nSet {}:max_render_file_size > {} to render.".format(
                f"{__virtualname__}.config", file_size
            )
        )
    return ret


def _format_markdown_link(name):
    link = name
    symbals = "~`!@#$%^&*()+={}[]:;\"<>,.?/|'\\"
    for s in symbals:
        link = link.replace(s, "")
    link = link.replace(" ", "-")
    return link


def _format_markdown_requisite(state, stateid, makelink=True):
    """
    format requisite as a link users can click
    """
    fmt_id = f"{state}: {stateid}"
    if makelink:
        return f" * [{fmt_id}](#{_format_markdown_link(fmt_id)})\n"
    else:
        return f" * `{fmt_id}`\n"


def processor_markdown(lowstate_item, config, **kwargs):
    """
    Takes low state data and returns a dict of processed data
    that is by default used in a jinja template when rendering a markdown highstate_doc.

    This `lowstate_item_markdown` given a lowstate item, returns a dict like:

    .. code-block:: none

        vars:       # the raw lowstate_item that was processed
        id:         # the 'id' of the state.
        id_full:    # combo of the state type and id "state: id"
        state:      # name of the salt state module
        function:   # name of the state function
        name:       # value of 'name:' passed to the salt state module
        state_function:    # the state name and function name
        markdown:          # text data to describe a state
            requisites:    # requisite like [watch_in, require_in]
            details:       # state name, parameters and other details like file contents

    """
    # TODO: switch or ... ext call.
    s = lowstate_item
    state_function = "{}.{}".format(s["state"], s["fun"])
    id_full = "{}: {}".format(s["state"], s["__id__"])

    # TODO: use salt defined STATE_REQUISITE_IN_KEYWORDS
    requisites = ""
    for comment, key in (
        ("run or update after changes in:\n", "watch"),
        ("after changes, run or update:\n", "watch_in"),
        ("require:\n", "require"),
        ("required in:\n", "require_in"),
    ):
        reqs = s.get(key, [])
        if reqs:
            requisites += comment
            for w in reqs:
                requisites += _format_markdown_requisite(*next(iter(w.items())))

    details = ""

    if state_function == "highstate_doc.note":
        if "contents" in s:
            details += "\n{}\n".format(s["contents"])
        if "source" in s:
            text = __salt__["cp.get_file_str"](s["source"])
            if text:
                details += f"\n{text}\n"
            else:
                details += "\n{}\n".format("ERROR: opening {}".format(s["source"]))

    if state_function == "pkg.installed":
        pkgs = s.get("pkgs", s.get("name"))
        details += f"\n```\ninstall: {pkgs}\n```\n"

    if state_function == "file.recurse":
        details += """recurse copy of files\n"""
        y = _state_data_to_yaml_string(s)
        if y:
            details += f"```\n{y}\n```\n"
        if "!doc_recurse" in id_full:
            findfiles = __salt__["file.find"](path=s.get("name"), type="f")
            if len(findfiles) < 10 or "!doc_recurse_force" in id_full:
                for f in findfiles:
                    details += _format_markdown_system_file(f, config)
            else:
                details += """ > Skipping because more than 10 files to display.\n"""
                details += (
                    """ > HINT: to force include !doc_recurse_force in state id.\n"""
                )
        else:
            details += """ > For more details review logs and Salt state files.\n\n"""
            details += """ > HINT: for improved docs use multiple file.managed states or file.archive, git.latest. etc.\n"""
            details += """ > HINT: to force doc to show all files in path add !doc_recurse .\n"""

    if state_function == "file.blockreplace":
        if s.get("content"):
            details += "ensure block of content is in file\n```\n{}\n```\n".format(
                _md_fix(s["content"])
            )
        if s.get("source"):
            text = "** source: " + s.get("source")
            details += "ensure block of content is in file\n```\n{}\n```\n".format(
                _md_fix(text)
            )

    if state_function == "file.managed":
        details += _format_markdown_system_file(s["name"], config)

    # if no state doc is created use default state as yaml
    if not details:
        y = _state_data_to_yaml_string(s)
        if y:
            details += f"```\n{y}```\n"

    r = {
        "vars": lowstate_item,
        "state": s["state"],
        "name": s["name"],
        "function": s["fun"],
        "id": s["__id__"],
        "id_full": id_full,
        "state_function": state_function,
        "markdown": {
            "requisites": requisites.decode("utf-8"),
            "details": details.decode("utf-8"),
        },
    }
    return r



import time

import pytest

from . import normalize_ret

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.core_test,
]


def test_requisites_full_sls_require(state, state_tree):
    """
    Test the sls special command in requisites
    """
    full_sls_contents = """
    B:
      cmd.run:
        - name: echo B
    C:
      cmd.run:
        - name: echo C
    """
    sls_contents = """
    include:
      - fullsls
    A:
      cmd.run:
        - name: echo A
        - require:
          - sls: fullsls
    """
    expected_result = {
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo A" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo B" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo C" run',
            "result": True,
            "changes": True,
        },
    }
    with pytest.helpers.temp_file(
        "requisite.sls", sls_contents, state_tree
    ), pytest.helpers.temp_file("fullsls.sls", full_sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_require_no_state_module(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # Complex require/require_in graph
    #
    # Relative order of C>E is given by the definition order
    #
    # D (1) <--+
    #          |
    # B (2) ---+ <-+ <-+ <-+
    #              |   |   |
    # C (3) <--+ --|---|---+
    #          |   |   |
    # E (4) ---|---|---+ <-+
    #          |   |       |
    # A (5) ---+ --+ ------+
    #

    A:
      cmd.run:
        - name: echo A fifth
        - require:
          - C
    B:
      cmd.run:
        - name: echo B second
        - require_in:
          - A
          - C

    C:
      cmd.run:
        - name: echo C third

    D:
      cmd.run:
        - name: echo D first
        - require_in:
          - B

    E:
      cmd.run:
        - name: echo E fourth
        - require:
          - B
        - require_in:
          - A

    # will fail with "The following requisites were not found"
    G:
      cmd.run:
        - name: echo G
        - require:
          - Z
    # will fail with "The following requisites were not found"
    H:
      cmd.run:
        - name: echo H
        - require:
          - Z
    """
    expected_result = {
        "cmd_|-A_|-echo A fifth_|-run": {
            "__run_num__": 4,
            "comment": 'Command "echo A fifth" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo B second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C third_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo C third" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-D_|-echo D first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo D first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-E_|-echo E fourth_|-run": {
            "__run_num__": 3,
            "comment": 'Command "echo E fourth" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-G_|-echo G_|-run": {
            "__run_num__": 5,
            "comment": "The following requisites were not found:\n"
            + "                   require:\n"
            + "                       id: Z\n",
            "result": False,
            "changes": False,
        },
        "cmd_|-H_|-echo H_|-run": {
            "__run_num__": 6,
            "comment": "The following requisites were not found:\n"
            + "                   require:\n"
            + "                       id: Z\n",
            "result": False,
            "changes": False,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_require_ordering_and_errors_1(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    expected_result = {
        "cmd_|-A_|-echo A fifth_|-run": {
            "__run_num__": 4,
            "comment": 'Command "echo A fifth" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B second_|-run": {
            "__run_num__": 1,
            "comment": 'Command "echo B second" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-echo C third_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo C third" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-D_|-echo D first_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo D first" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-E_|-echo E fourth_|-run": {
            "__run_num__": 3,
            "comment": 'Command "echo E fourth" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-F_|-echo F_|-run": {
            "__run_num__": 5,
            "comment": "The following requisites were not found:\n"
            + "                   require:\n"
            + "                       foobar: A\n",
            "result": False,
            "changes": False,
        },
        "cmd_|-G_|-echo G_|-run": {
            "__run_num__": 6,
            "comment": "The following requisites were not found:\n"
            + "                   require:\n"
            + "                       cmd: Z\n",
            "result": False,
            "changes": False,
        },
        "cmd_|-H_|-echo H_|-run": {
            "__run_num__": 7,
            "comment": "The following requisites were not found:\n"
            + "                   require:\n"
            + "                       cmd: Z\n",
            "result": False,
            "changes": False,
        },
    }
    sls_contents = """
    # Complex require/require_in graph
    #
    # Relative order of C>E is given by the definition order
    #
    # D (1) <--+
    #          |
    # B (2) ---+ <-+ <-+ <-+
    #              |   |   |
    # C (3) <--+ --|---|---+
    #          |   |   |
    # E (4) ---|---|---+ <-+
    #          |   |       |
    # A (5) ---+ --+ ------+
    #

    A:
      cmd.run:
        - name: echo A fifth
        - require:
          - cmd: C
    B:
      cmd.run:
        - name: echo B second
        - require_in:
          - cmd: A
          - cmd: C

    C:
      cmd.run:
        - name: echo C third

    D:
      cmd.run:
        - name: echo D first
        - require_in:
          - cmd: B

    E:
      cmd.run:
        - name: echo E fourth
        - require:
          - cmd: B
        - require_in:
          - cmd: A

    # will fail with "The following requisites were not found"
    F:
      cmd.run:
        - name: echo F
        - require:
          - foobar: A
    # will fail with "The following requisites were not found"
    G:
      cmd.run:
        - name: echo G
        - require:
          - cmd: Z
    # will fail with "The following requisites were not found"
    H:
      cmd.run:
        - name: echo H
        - require:
          - cmd: Z
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_require_ordering_and_errors_2(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # will fail with "Data failed to compile:"
    A:
      cmd.run:
        - name: echo A
        - require_in:
          - foobar: W
    """
    errmsg = (
        "Cannot extend ID 'W' in 'base:requisite'. It is not part of the high"
        " state.\nThis is likely due to a missing include statement or an incorrectly"
        " typed ID.\nEnsure that a state with an ID of 'W' is available\nin environment"
        " 'base' and to SLS 'requisite'"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_requisites_require_ordering_and_errors_3(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # issue #8772
    # should fail with "Data failed to compile:"
    B:
      cmd.run:
        - name: echo B last
        - require_in:
          # state foobar does not exists in A
          - foobar: A

    A:
      cmd.run:
        - name: echo A first
    """
    errmsg = (
        "Cannot extend ID 'A' in 'base:requisite'. It is not part of the high state.\n"
        "This is likely due to a missing include statement or an incorrectly typed ID.\n"
        "Ensure that a state with an ID of 'A' is available\n"
        "in environment 'base' and to SLS 'requisite'"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_requisites_require_ordering_and_errors_4(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    A:
      cmd.run:
        - name: echo A
    B:
      cmd.run:
        - name: echo B
        # here used without "-"
        - require:
            cmd: A
    C:
      cmd.run:
        - name: echo C
        # here used without "-"
        - require_in:
            cmd: A
    """
    # issue #8235
    # FIXME: Why is require enforcing list syntax while require_in does not?
    # And why preventing it?
    # Currently this state fails, should return C/B/A
    errmsg = (
        "The require statement in state 'B' in SLS 'requisite' needs to be formed as a"
        " list"
    )
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_requisites_require_ordering_and_errors_5(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    A:
      cmd.run:
        - name: echo A
        - require:
          - cmd: B

    B:
      cmd.run:
        - name: echo B
        - require:
          - cmd: A
    """
    # issue #8235
    # FIXME: Why is require enforcing list syntax while require_in does not?
    # And why preventing it?
    # Currently this state fails, should return C/B/A
    errmsg = 'A recursive requisite was found, SLS "requisite" ID "B" ID "A"'
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert ret.failed
        assert ret.errors == [errmsg]


def test_requisites_require_any(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # Complex require/require_in graph
    #
    # Relative order of C>E is given by the definition order
    #
    # D (1) <--+
    #          |
    # B (2) ---+ <-+ <-+ <-+
    #              |   |   |
    # C (3) <--+ --|---|---+
    #          |   |   |
    # E (4) ---|---|---+ <-+
    #          |   |       |
    # A (5) ---+ --+ ------+
    #

    # A should success since B succeeds even though C fails.
    A:
      cmd.run:
        - name: echo A
        - require_any:
          - cmd: B
          - cmd: C
          - cmd: D
    B:
      cmd.run:
        - name: echo B

    C:
      cmd.run:
        - name: "$(which false)"

    D:
      cmd.run:
        - name: echo D
    """
    expected_result = {
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 3,
            "comment": 'Command "echo A" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-B_|-echo B_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo B" run',
            "result": True,
            "changes": True,
        },
        "cmd_|-C_|-$(which false)_|-run": {
            "__run_num__": 1,
            "comment": 'Command "$(which false)" run',
            "result": False,
            "changes": True,
        },
        "cmd_|-D_|-echo D_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo D" run',
            "result": True,
            "changes": True,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_requisites_require_any_fail(state, state_tree):
    """
    Call sls file containing several require_in and require.

    Ensure that some of them are failing and that the order is right.
    """
    sls_contents = """
    # D should fail since both E & F fail
    E:
      cmd.run:
        - name: 'false'

    F:
      cmd.run:
        - name: 'false'

    D:
      cmd.run:
        - name: echo D
        - require_any:
          - cmd: E
          - cmd: F
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert "One or more requisite failed" in ret["cmd_|-D_|-echo D_|-run"].comment


def test_issue_38683_require_order_failhard_combination(state, state_tree):
    """
    This tests the case where require, order, and failhard are all used together in a state definition.

    Previously, the order option, which used in tandem with require and failhard, would cause the state
    compiler to stacktrace. This exposed a logic error in the ``check_failhard`` function of the state
    compiler. With the logic error resolved, this test should now pass.

    See https://github.com/saltstack/salt/issues/38683 for more information.
    """
    sls_contents = """
    a:
      test.show_notification:
        - name: a
        - text: message
        - require:
            - test: b
        - order: 1
        - failhard: True

    b:
      test.fail_with_changes:
        - name: b
        - failhard: True
    """
    state_id = "test_|-b_|-b_|-fail_with_changes"
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        assert state_id in ret
        assert ret[state_id].result is False
        assert ret[state_id].comment == "Failure!"


@pytest.mark.skip_on_windows
def test_parallel_state_with_requires(state, state_tree):
    """
    This is a test case for https://github.com/saltstack/salt/issues/49273
    Parallel state object has any requisites
    """
    sls_contents = """
    barrier:
      cmd.run:
        - name: sleep 1

    {%- for x in range(1, 10) %}
    blah-{{x}}:
      cmd.run:
        - name: sleep 2
        - require:
          - barrier
          - barrier2
        - parallel: true
    {% endfor %}

    barrier2:
      test.nop
    """
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        start_time = time.time()
        ret = state.sls(
            "requisite",
            __pub_jid="1",  # Because these run in parallel we need a fake JID)
        )
        end_time = time.time()

        # We're running 3 states that sleep for 10 seconds each
        # they'll run in parallel so we should be below 30 seconds
        # confirm that the total runtime is below 30s
        assert (end_time - start_time) < 30

        for item in range(1, 10):
            _id = f"cmd_|-blah-{item}_|-sleep 2_|-run"
            assert "__parallel__" in ret[_id]


def test_issue_59922_conflict_in_name_and_id_for_require_in(state, state_tree):
    """
    Make sure that state_type is always honored while compiling down require_in to

    corresponding require statement.
    """
    sls_contents = """
    X:
      test.succeed_without_changes:
        - name: A

    A:
      cmd.run:
        - name: echo A

    B:
      cmd.run:
        - name: echo B
        - require_in:
          - test: A
    """
    expected_result = {
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 2,
            "comment": 'Command "echo A" run',
            "result": True,
            "changes": True,
        },
        "test_|-X_|-A_|-succeed_without_changes": {
            "__run_num__": 1,
            "comment": "Success!",
            "result": True,
            "changes": False,
        },
        "cmd_|-B_|-echo B_|-run": {
            "__run_num__": 0,
            "comment": 'Command "echo B" run',
            "result": True,
            "changes": True,
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result


def test_issue_61121_extend_is_to_strict(state, state_tree):
    """
    test that extend works as advertised with adding new service_types to
    a state id
    """

    sls_contents = """
    A:
      test.succeed_without_changes:
        - name: a
    extend:
      A:
        cmd:
          - run
          - name: echo A
    """
    expected_result = {
        "test_|-A_|-a_|-succeed_without_changes": {
            "__run_num__": 0,
            "changes": False,
            "result": True,
            "comment": "Success!",
        },
        "cmd_|-A_|-echo A_|-run": {
            "__run_num__": 1,
            "changes": True,
            "result": True,
            "comment": 'Command "echo A" run',
        },
    }
    with pytest.helpers.temp_file("requisite.sls", sls_contents, state_tree):
        ret = state.sls("requisite")
        result = normalize_ret(ret.raw)
        assert result == expected_result



"""
    :codeauthor: Jochen Breuer <jbreuer@suse.de>
"""

# pylint: disable=no-value-for-parameter
import os
from contextlib import contextmanager

import pytest

import salt.utils.files
import salt.utils.platform
from salt.modules import kubernetesmod as kubernetes
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import Mock, patch
from tests.support.unit import TestCase


@contextmanager
def mock_kubernetes_library():
    """
    After fixing the bug in 1c821c0e77de58892c77d8e55386fac25e518c31,
    it caused kubernetes._cleanup() to get called for virtually every
    test, which blows up. This prevents that specific blow-up once
    """
    with patch("salt.modules.kubernetesmod.kubernetes") as mock_kubernetes_lib:
        yield mock_kubernetes_lib


@pytest.mark.skipif(
    not kubernetes.HAS_LIBS,
    reason="Kubernetes client lib is not installed. Skipping test_kubernetes.py",
)
class KubernetesTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for salt.modules.kubernetesmod
    """

    def setup_loader_modules(self):
        return {kubernetes: {"__salt__": {}}}

    def test_nodes(self):
        """
        Test node listing.
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.client.CoreV1Api.return_value = Mock(
                    **{
                        "list_node.return_value.to_dict.return_value": {
                            "items": [{"metadata": {"name": "mock_node_name"}}]
                        }
                    }
                )
                self.assertEqual(kubernetes.nodes(), ["mock_node_name"])
                self.assertTrue(
                    kubernetes.kubernetes.client.CoreV1Api().list_node().to_dict.called
                )

    def test_deployments(self):
        """
        Tests deployment listing.
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.client.ExtensionsV1beta1Api.return_value = Mock(
                    **{
                        "list_namespaced_deployment.return_value.to_dict.return_value": {
                            "items": [{"metadata": {"name": "mock_deployment_name"}}]
                        }
                    }
                )
                self.assertEqual(kubernetes.deployments(), ["mock_deployment_name"])
                # pylint: disable=E1120
                self.assertTrue(
                    kubernetes.kubernetes.client.ExtensionsV1beta1Api()
                    .list_namespaced_deployment()
                    .to_dict.called
                )
                # pylint: enable=E1120

    def test_services(self):
        """
        Tests services listing.
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.client.CoreV1Api.return_value = Mock(
                    **{
                        "list_namespaced_service.return_value.to_dict.return_value": {
                            "items": [{"metadata": {"name": "mock_service_name"}}]
                        }
                    }
                )
                self.assertEqual(kubernetes.services(), ["mock_service_name"])
                # pylint: disable=E1120
                self.assertTrue(
                    kubernetes.kubernetes.client.CoreV1Api()
                    .list_namespaced_service()
                    .to_dict.called
                )
                # pylint: enable=E1120

    def test_pods(self):
        """
        Tests pods listing.
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.client.CoreV1Api.return_value = Mock(
                    **{
                        "list_namespaced_pod.return_value.to_dict.return_value": {
                            "items": [{"metadata": {"name": "mock_pod_name"}}]
                        }
                    }
                )
                self.assertEqual(kubernetes.pods(), ["mock_pod_name"])
                # pylint: disable=E1120
                self.assertTrue(
                    kubernetes.kubernetes.client.CoreV1Api()
                    .list_namespaced_pod()
                    .to_dict.called
                )
                # pylint: enable=E1120

    def test_delete_deployments(self):
        """
        Tests deployment deletion
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch(
                "salt.modules.kubernetesmod.show_deployment", Mock(return_value=None)
            ):
                with patch.dict(
                    kubernetes.__salt__,
                    {"config.option": Mock(side_effect=self.settings)},
                ):
                    mock_kubernetes_lib.client.V1DeleteOptions = Mock(return_value="")
                    mock_kubernetes_lib.client.ExtensionsV1beta1Api.return_value = Mock(
                        **{
                            "delete_namespaced_deployment.return_value.to_dict.return_value": {
                                "code": ""
                            }
                        }
                    )
                    self.assertEqual(
                        kubernetes.delete_deployment("test"), {"code": 200}
                    )
                    # pylint: disable=E1120
                    self.assertTrue(
                        kubernetes.kubernetes.client.ExtensionsV1beta1Api()
                        .delete_namespaced_deployment()
                        .to_dict.called
                    )
                    # pylint: enable=E1120

    def test_create_deployments(self):
        """
        Tests deployment creation.
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.client.ExtensionsV1beta1Api.return_value = Mock(
                    **{
                        "create_namespaced_deployment.return_value.to_dict.return_value": {}
                    }
                )
                self.assertEqual(
                    kubernetes.create_deployment(
                        "test", "default", {}, {}, None, None, None
                    ),
                    {},
                )
                # pylint: disable=E1120
                self.assertTrue(
                    kubernetes.kubernetes.client.ExtensionsV1beta1Api()
                    .create_namespaced_deployment()
                    .to_dict.called
                )
                # pylint: enable=E1120

    @staticmethod
    def settings(name, value=None):
        """
        Test helper
        :return: settings or default
        """
        data = {
            "kubernetes.kubeconfig": "/home/testuser/.minikube/kubeconfig.cfg",
            "kubernetes.context": "minikube",
        }
        return data.get(name, value)

    def test_setup_kubeconfig_file(self):
        """
        Test that the `kubernetes.kubeconfig` configuration isn't overwritten
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.config.load_kube_config = Mock()
                config = kubernetes._setup_conn()
                self.assertEqual(
                    self.settings("kubernetes.kubeconfig"),
                    config["kubeconfig"],
                )

    def test_setup_kubeconfig_data_overwrite(self):
        """
        Test that provided `kubernetes.kubeconfig` configuration is overwritten
        by provided kubeconfig_data in the command
        :return:
        """
        with mock_kubernetes_library() as mock_kubernetes_lib:
            with patch.dict(
                kubernetes.__salt__, {"config.option": Mock(side_effect=self.settings)}
            ):
                mock_kubernetes_lib.config.load_kube_config = Mock()
                config = kubernetes._setup_conn(
                    kubeconfig_data="MTIzNDU2Nzg5MAo=", context="newcontext"
                )
                check_path = os.path.join("/tmp", "salt-kubeconfig-")
                if salt.utils.platform.is_windows():
                    check_path = os.path.join(os.environ.get("TMP"), "salt-kubeconfig-")
                elif salt.utils.platform.is_darwin():
                    check_path = os.path.join(
                        os.environ.get("TMPDIR", "/tmp"), "salt-kubeconfig-"
                    )
                self.assertTrue(
                    config["kubeconfig"].lower().startswith(check_path.lower())
                )
                self.assertTrue(os.path.exists(config["kubeconfig"]))
                with salt.utils.files.fopen(config["kubeconfig"], "r") as kcfg:
                    self.assertEqual("1234567890\n", kcfg.read())
                kubernetes._cleanup(**config)

    def test_node_labels(self):
        """
        Test kubernetes.node_labels
        :return:
        """
        with patch("salt.modules.kubernetesmod.node") as mock_node:
            mock_node.return_value = {
                "metadata": {
                    "labels": {
                        "kubernetes.io/hostname": "minikube",
                        "kubernetes.io/os": "linux",
                    }
                }
            }
            self.assertEqual(
                kubernetes.node_labels("minikube"),
                {"kubernetes.io/hostname": "minikube", "kubernetes.io/os": "linux"},
            )

    def test_adding_change_cause_annotation(self):
        """
        Tests adding a `kubernetes.io/change-cause` annotation just like
        kubectl [apply|create|replace] --record does
        :return:
        """
        with patch(
            "salt.modules.kubernetesmod.sys.argv", ["/usr/bin/salt-call", "state.apply"]
        ) as mock_sys:
            func = getattr(kubernetes, "__dict_to_object_meta")
            data = func(name="test-pod", namespace="test", metadata={})

            self.assertEqual(data.name, "test-pod")
            self.assertEqual(data.namespace, "test")
            self.assertEqual(
                data.annotations,
                {"kubernetes.io/change-cause": "/usr/bin/salt-call state.apply"},
            )

            # Ensure any specified annotations aren't overwritten
            test_metadata = {"annotations": {"kubernetes.io/change-cause": "NOPE"}}
            data = func(name="test-pod", namespace="test", metadata=test_metadata)

            self.assertEqual(data.annotations, {"kubernetes.io/change-cause": "NOPE"})

    def test_enforce_only_strings_dict(self):
        func = getattr(kubernetes, "__enforce_only_strings_dict")
        data = {
            "unicode": 1,
            2: 2,
        }
        self.assertEqual(
            {"unicode": "1", "2": "2"},
            func(data),
        )



"""
    :codeauthor: :email:`Simon Dodsley <simon@purestorage.com>`
"""

import errno
import textwrap

import salt.grains.nvme as nvme
from tests.support.mock import mock_open, patch


def test_linux_nvme_nqn_grains():
    _nvme_file = textwrap.dedent(
        """\
        nqn.2014-08.org.nvmexpress:fc_lif:uuid:2cd61a74-17f9-4c22-b350-3020020c458d
        """
    )

    with patch("salt.utils.files.fopen", mock_open(read_data=_nvme_file)):
        nqn = nvme._linux_nqn()

    assert isinstance(nqn, list)
    assert len(nqn) == 1
    assert nqn == [
        "nqn.2014-08.org.nvmexpress:fc_lif:uuid:2cd61a74-17f9-4c22-b350-3020020c458d"
    ]


def test_linux_nqn_non_root():
    """
    Test if linux_nqn is running on salt-master as non-root
    and handling access denial properly.
    :return:
    """
    with patch(
        "salt.utils.files.fopen",
        side_effect=IOError(errno.EPERM, "The cables are not the same length."),
    ):
        with patch("salt.grains.nvme.log"):
            assert nvme._linux_nqn() == []
            nvme.log.debug.assert_called()
            assert "Error while accessing" in nvme.log.debug.call_args[0][0]
            assert "cables are not the same" in nvme.log.debug.call_args[0][2].strerror
            assert nvme.log.debug.call_args[0][2].errno == errno.EPERM
            assert nvme.log.debug.call_args[0][1] == "/etc/nvme/hostnqn"


def test_linux_nqn_no_nvme_initiator():
    """
    Test if linux_nqn is running on salt-master as root.
    nvme initiator is not there accessible or is not supported.
    :return:
    """
    with patch("salt.utils.files.fopen", side_effect=IOError(errno.ENOENT, "")):
        with patch("salt.grains.nvme.log"):
            assert nvme._linux_nqn() == []
            nvme.log.debug.assert_not_called()



import logging

log = logging.getLogger(__name__)


def test_noop_return(caplog, salt_cli, salt_minion):
    # This test relies on the customer noop returner found at
    # tests/integartion/files/returners/noop_returner.py. The path to the
    # returner must be in the master's config and 'rutests_noop' should be in
    # the master's event_return setting. The master's log level must also be
    # set to DEBUG. These config settings are found in
    # tests/pytests/conftest.py::master_factory
    with caplog.at_level(logging.DEBUG):
        salt_cli.run("test.ping", minion_tgt=salt_minion.id)
        assert "NOOP_RETURN" in caplog.text



import pathlib

import salt.utils.job


def test_store_job_save_load(minion_opts, tmp_path):
    """
    Test to ensure we create the correct files when we
    store a job in the local cache
    """
    opts = minion_opts.copy()
    opts["master_job_cache"] = "local_cache"
    opts["job_cache"] = True
    opts["ext_job_cache"] = ""
    cache_dir = pathlib.Path(opts["cachedir"], "jobs")
    master_minion = "test_master"
    load = {
        "id": master_minion,
        "tgt": master_minion,
        "jid": "20230822145508520090",
        "return": {
            "fun": "runner.test.arg",
            "jid": "20230822145508520090",
            "user": "sudo_ch3ll",
            "fun_args": ["go", "home"],
            "_stamp": "2023-08-22T14:55:08.796680",
            "return": {"args": ("go", "home"), "kwargs": {}},
            "success": True,
        },
    }
    salt.utils.job.store_job(opts, load)
    job_dir = list(list(cache_dir.iterdir())[0].iterdir())[0]
    return_p = job_dir / master_minion / "return.p"
    load_p = job_dir / ".load.p"
    jid = job_dir / "jid"
    assert return_p.is_file()
    assert load_p.is_file()
    assert jid.is_file()
